function [hist, figh] = clreplay(model, cfg)
% Replays an arbitrary policy with a system, records & plots the results
%   [HIST, FIGH] = CLREPLAY(MODEL, CFG)
%   [HIST, FIGH] = CLREPLAY(PROBLEM, CFG)
%
% Parameters:
%   PROBLEM     - the problem function (preferably a function handle) 
%       or
%   MODEL       - an existing model structure
%   CFG         - configuration, see commented defaults. A control policy can be
%       specified as a configuration field.
% Returns:
%   HIST        - the history
%   FIGH        - handle of the figure containing the plotted trajectories

% Default config
CFG.gamma = [];             % discount factor
CFG.tend = 30;              % max length of simulation in time units (not steps)
CFG.x0 = [];                % initial state; problem default or 0 if not specified
CFG.modelargs = {};         % if only problem specified, this holds problem args in 'model' mode
CFG.policy = '';            % policy for control
CFG.policyargs = {};        % its arguments
CFG.plotfuns = {};          % variable-wise plot functions
CFG.noplot = 0;             % disable plot trajectory
% figure output config
CFG = setfigprop(CFG, 'addfields');

% cfg

% process inputs
if nargin >= 2, cfg = str2cfg(cfg, CFG); cfg = checkparams(cfg, CFG); 
else cfg = CFG; end;
cfg.grayscale = grayscalefromconfig(cfg);

% create model if not given
if ~isstruct(model),
    problem = model;
    model = feval(problem, 'model', cfg.modelargs{:});
else
    problem = '';
end;    

% initial state
if ~isempty(cfg.x0), 
    if ischar(cfg.x0) && ~isempty(problem),
        % use the model function to get a string-named initial state
        x0 = feval(problem, 'x0', cfg.x0);
    else
        x0 = cfg.x0(:);         % specified initial state
    end;
else
    % try getting default initial state
    try
        x0 = feval(problem, 'x0');
    catch
        x0 = zeros(model.p, 1);    % zeros
    end;
end;         

% init history
maxsteps = cfg.tend / model.Ts;
t = (0 : maxsteps) * model.Ts;
Ns = length(t) - 1;
x = zeros(model.p, length(t)); 
u = zeros(model.q, length(t));
u(:, end) = NaN;
r = zeros(1, length(t)); r(1) = NaN;

if ~isempty(cfg.policy),
    % determine the order of the possibly dynamic controller
    try
        [cx0, cord] = feval(cfg.policy, 'init', cfg.policyargs{:});
    catch
        cord = 0;   % static feedback -- or old-style policy, which is always static
    end;
end;
if cord > 0,    % init controller state history
    cx = zeros(cord, length(t));
    cx(:, 1) = cx0;
end;

% samples loop
tmark = cputime;
x(:, 1) = x0;
for k = 1:Ns,
    if ~isempty(cfg.policy),        % use controller to compute action
        if cord > 0,    % dynamic controller
            [cx(:, k+1), u(:, k)] = feval(cfg.policy, cx(:, k), x(:, k), model.Ts, k, cfg.policyargs{:});
        else            % static controller
            u(:, k) = feval(cfg.policy, x(:, k), (k-1) * model.Ts, cfg.policyargs{:});
        end;
    else                            % no control
        u(:, k) = 0;
    end;
    % apply to system
    [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
    if terminal, Ns = k; break; end;      % entered terminal state
end;
trun = cputime - tmark;     % runtime for simulation (currently unused)

% form and save history
hist.Ns = Ns;
hist.R = discreturn(cfg, r, Ns, terminal);
hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);    
% add controller state if controller is dynamic
if cord > 0,
    hist.cx = cx(:, 1:Ns+1);
end;

% plot if enabled in config
if ~cfg.noplot,
    % toscreen = strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget);
    if isfield(model, 'plotfun'),
        pcfg = struct; pcfg.traj = hist; pcfg.hist = hist;
        pcfg = copyfields(cfg, pcfg, fieldnames(setfigprop(struct, 'addfields')));
        figh = feval(model.plotfun, pcfg);
        % figh = feval(model.plotfun, hist); % old style
    else
        figh = plothistory(hist);
    end;
    setfigprop(cfg);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% end ODETEST() returning EVOL, TIMES, FIGH