% Demonstration of DP/RL algorithms on the inverted pendulum problem

%% Initialization

% common config
ccfg.target = 'screen';      % set to "files" to output figures and movies; to "screen" for live demo
% ccfg.target = 'files';      % set to "files" to output figures and movies; to "screen" for live demo
ccfg.problem = 'ipsetup_problem';
ccfg.gamma = 0.98;
ccfg.model_params = {[], 'type=ip maxx=[pi;15*pi] maxu=3', 'Ts=0.005'};
ccfg.N = [41 21];
ccfg.M = 5;
ccfg.eps = 1;
ccfg.datadir = pwd;
% visualizer config
ccfg.viscfg = struct;
ccfg.viscfg.figsize = [800 500];
ccfg.viscfg.pause = -1;

%% 1) Fuzzy Q-iteration
cfg = ccfg;
cfg.xgrids = symequidgrid(cfg.N, [pi, 15*pi]);
cfg.ugrids = {symequidgrid(cfg.M, 3)};
cfg.run = 1;
cfg.savetheta = 1;
cfg.datafile = 'ipsetup_fuzzyqidemo';
%
cfg.visualize = 1;
cfg.iterdisp = 5;
cfg.itervis = 10;
fuzzyqi(cfg);

% cleanup
% delete([cfg.datafile '.mat']);


%% 2) LLR Fitted QI
cfg = ccfg;
cfg.samples.method = 'grid';
cfg.samples.xgrids = symequidgrid([31 15], [pi, 15*pi]);
cfg.samples.ugrids = {symequidgrid(3, 3)};
cfg.U = cfg.samples.ugrids;
cfg.knn = 5;
cfg.maxiter = 500;
cfg.run = 1;
cfg.datafile = 'ipsetup_llrfittedqidemo';
%
cfg.visualize = 1;
cfg.iterdisp = 1;
cfg.itervis = 20;
cfg.viscfg.savefig = 'ipsetup_llrfittedqidemo';
cfg.viscfg.snapshot = 130;
cfg.viscfg.figsize = [600 300];
% rand('twister', 1000);  % uncomment for repeatable results
llrfittedqi(cfg);

% cleanup
delete([cfg.datafile '.mat']);

%% 2) Fitted QI with extra-trees
% Requires 32-bit Matlab to run the mexfile for extra-trees (rtenslearn_c)
cfg = ccfg;
cfg.samples.method = 'grid';
cfg.samples.xgrids = symequidgrid([31 15], [pi, 15*pi]);
cfg.samples.ugrids = {symequidgrid(3, 3)};
cfg.U = cfg.samples.ugrids;
cfg.singlereg = 1;
cfg.maxiter = 100;
cfg.run = 1;
cfg.datafile = 'ipsetup_fittedqidemo';
%
cfg.visualize = 1;
cfg.iterdisp = 1;
cfg.itervis = 10;
cfg.viscfg.savefig = 'ipsetup_fittedqidemo';
cfg.viscfg.snapshot = 50;
cfg.viscfg.figsize = [600 300];
% rand('twister', 1000);  % uncomment for repeatable results
fittedqi(cfg);

% cleanup
delete([cfg.datafile '.mat']);
  

%% 3) LSPI
cfg = ccfg;
cfg.approx = 'enhanced type=rbfdisc N=[11,15] M=3 xspace=eq uspace=eq';
cfg.samples = 'randdisc N=10000';
cfg.run = 1;
cfg.datafile = 'ipsetup_lspidemo';
%
cfg.visualize = 1;
cfg.iterdisp = 1;
cfg.itervis = 1;
cfg.viscfg.savefig = 'ipsetup_lspidemo';            
cfg.viscfg.snapshot = 3;
% rand('twister', 1000);  % uncomment for repeatable results
lspi(cfg);

% cleanup
% delete([cfg.datafile '.mat']);



%% 4) Optimistic planning

% algorithm parameters
cfg = ccfg;
% we also change the reward function to so-called "normalized LQR" -- the same as before, except now
% it's scaled and translated into [0,1]
cfg.model_params = {'nlqr', 'type=ip maxx=[pi;15*pi] maxu=3', 'Ts=0.05'};
% planning algorithm parameters
cfg.planalg = 'opd';            % variant: optimistic planning for deterministic systems
cfg.planparams.n = 50;         % budget of expansions at each step
cfg.planparams.U = [-3 0 3];    % action space -- the same as for M=3 in the approximators above
cfg.reset = [-pi, 0]';          % always reset pointing down
cfg.trials = 1;                 % one trial is enough to see performance, as we're not learning
cfg.maxsteps = 80;              % 80 steps = 4 seconds should be enough for a swingup
% output config
cfg.datadir = pwd;
cfg.datafile = 'ipsetup_optplan';
% use visualization -- set visualize to 0 to disable
cfg.visualize = 3;             % this means: only visualize system steps
cfg.itervis = 1;
cfg.viscfg.pause = 0.01;      

% run the algorithm
cfg.run = 1;
genmpc(cfg);              

% look at the controlled trajectory
genmpc('plottrial datafile=ipsetup_optplan');

%% 5) OPSS

% algorithm parameters
cfg = ccfg;
% we also change the reward function to so-called "normalized LQR" -- the same as before, except now
% it's scaled and translated into [0,1]
cfg.model_params = {'nlqr', 'type=ip maxx=[pi;15*pi] maxu=3 stoch_prob=[.6,.4] stoch_gain=[1,.7]', 'Ts=0.05'};
% planning algorithm parameters
cfg.planalg = 'opss';            % variant: optimistic planning for deterministic systems
cfg.planparams.n = 150;          % budget of expansions at each step
cfg.planparams.U = [-3 0 3];     % action space -- the same as for M=3 in the approximators above
% reuse subtree from previous iterations
cfg.planparams.reusetree = 1;    
% uncomment these to learn the b-function (but then reusetree must be 0)
% cfg.planparams.usemem = 1;
% cfg.planparams.lipschitz = 0.5;
% cfg.planparams.mem_maxsize = 100;
cfg.reset = [-pi, 0]';          % always reset pointing down
cfg.trials = 1;                 % one trial is enough to see performance, as we're not learning
cfg.maxsteps = 50;              % 80 steps = 4 seconds should be enough for a swingup
% output config
cfg.datadir = pwd;
cfg.datafile = 'ipsetupstoch_opss';
% use visualization -- set visualize to 0 to disable
cfg.visualize = 3;             % this means: only visualize system steps
cfg.viscfg.pause = 0.01;      

% run the algorithm
cfg.run = 1;
genmpc(cfg);              

% look at the controlled trajectory
genmpc('plottrial datafile=ipsetupstoch_opss');
