% LSPI_DEMO Demonstration of the use of the LSPI functions in approximate RL package
% The demo is quite CPU-intensive and takes a few minutes to complete!

%% =====================================
% Let's first configure the common settings for the algorithms

cfg = struct;
cfg.problem = 'ipsetup_problem';
% we configure the model such that it creates a swingup problem:
% control bound is 3 Volts, weight is added. The sampling time is 5 thousands.
cfg.model_params = {[], 'type=ip maxu=3', 'Ts=0.005'};
cfg.gamma = .98;        % discount factor
cfg.eps = 0.1;          % convergence threshold
cfg.maxiter  = 25;      % max # iterations: make it quick
% save all the datafiles in the current directory
cfg.datadir = pwd;

%% =====================================
% Demonstrate the use of the basic algorithms

lcfg = cfg;             % Start from the common config

% Q-function approximator configuration
% we use "enhanced" mode and set a grid of equally spaced RBFs
% we use 3 discrete actions
lcfg.approx = 'enhanced type=rbfdisc N=11 M=3 xspace=eq';
% lcfg.approx.enhanced = 0;
% lcfg.approx.c = flat({[-2, -1, 0, 3], [-1, 0, 1]});
% lcfg.approx.rad = repmat([0.3; 0.2], 1, size(lcfg.approx.c, 2));
% lcfg.approx.ugrids = {[-3, 0, 3], [-1, 0, 1]};
% samples for policy evaluation; note we use discrete samples for the actions
lcfg.samples = 'method=randdisc N=7500';

%% =====================================
% The offline LSPI algorithm

% delete data file for the purposes of the demo
if exist('demo_lspidata.mat', 'file'), delete('demo_lspidata.mat'); end;

% Run the basic LSPI algorithm
lspicfg = lcfg;         % start from the basic algorithm config
lspicfg.run = 1;        % we want to run the algorithm
lspicfg.datafile = 'demo_lspidata'; % save to this data file
lspi(lspicfg);          % call the function

% Examine the solution
lspicfg = lcfg;
lspicfg.sol = 1;
lspicfg.datafile = 'demo_lspidata';
lspi(lspicfg);
% It looks reasonable

% Let's see how it does controlling the system
lspicfg = lcfg;
lspicfg.replay = 1;
lspicfg.datafile = 'demo_lspidata';
lspicfg.x0 = [-pi,0]';  % start pointing down
lspicfg.tend = 3;       % should be sufficient to stabilize upwards
lspi(lspicfg);

%% =====================================
% The online LSPI algorithm

% delete data file for the purposes of the demo
if exist('demo_lspionldata.mat', 'file'), delete('demo_lspionldata.mat'); end;

% Run the basic LSPI algorithm
lspiocfg = lcfg;            % start from the basic algorithm config
lspiocfg.run = 1;           % we want to run the algorithm
lspiocfg.maxtime = 120;     % run for two simulated minutes
lspiocfg.trialtime = 1.5;   % 1.5 seconds per trial
lspiocfg.reset = 'rand';    % reset to random states after every trial
lspiocfg.dupd = 30 * 0.005; % update each 10 samples
lspiocfg.ddisp = 10;        % display stats after each 10 seconds of simulated time
% decay exploration so that in the end it's 0.05
lspiocfg.explordecay = 0.05^(1/lspiocfg.maxtime);
lspiocfg.datafile = 'demo_lspionldata'; % save to this data file
lspionline(lspiocfg);          % call the function

% Examine the solution
lspiocfg = lcfg;
lspiocfg.sol = 1;
lspiocfg.datafile = 'demo_lspionldata';
lspionline(lspiocfg);
% It looks reasonable

% Look at some trajectories during learning
lspiocfg = lcfg;
lspiocfg.traj = 1;
lspiocfg.dtraj = 10;    % a trajectory once every 10 seconds
lspiocfg.datafile = 'demo_lspionldata';
lspionline(lspiocfg);

% Let's see how it does controlling the system
lspiocfg = lcfg;
lspiocfg.replay = 1;
lspiocfg.datafile = 'demo_lspionldata';
lspiocfg.x0 = [-pi,0]';  % start pointing down
lspiocfg.tend = 3;       % should be sufficient to stabilize upwards
lspionline(lspiocfg);


% % =====================================
% % Batch experiments demo
% 
% bcfg = cfg;             % Start from the common config
% % Note all the common config fields are taken into account and properly transmitted to the
% % individual LSPI calls
% 
% % Say we want to see how changing the # of basis functions influences the solution
% bcfg.alg = 'lspi';      % the basic LSPI algorithm is studied
% bcfg.type = 'rbfdisc';
% bcfg.N = [9 11];        % we use 9x9 and 11x11 basis functions
% bcfg.M = 3;             % always use 3 discrete actions
% bcfg.method = 'randdisc';
% bcfg.Ns = 6000;         % this is the number of samples for N = 9
% bcfg.adaptiveNs = 1;    % adapt it proportionally to # of parameters
% bcfg.infix = 'demo';
% bcfg.repeat = 1;        % run each experiment only once
% % performance evaluation: use a small set of initial samples for speed
% bcfg.X0 = 'small'; 
% bcfg.mc_eps = 0.1;       % accept an error of at most 0.1
% 
% % Run the batch
% ecfg = bcfg;
% ecfg.run = 1;
% lspiexp(ecfg);
% 
% % Let's plot the performance and # of iterations to convergence for each exp
% ecfg = bcfg;
% ecfg.plot = {'perf', 'iter'};
% ecfg.stats = 1;      % make sure performance is evaluated, it's needed in the plots
% % change the location of the cache so this works on any computer
% ecfg.pcache = [pwd '\' 'demo_cache'];
% lspiexp(ecfg);

% close all; echo off; return; 
