function  [uroot, planinfo, plandata] = maopd_optimizeneighbors(xroot, model, cfg)
% OPD Optimistic planning with deterministic policies for multiagent systems
% Optimizes the actions of all neighbors (using either maxmin or maxmax solution)

% Uses arrays; does not support tree or memory reuse
% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.gametype = 'max';           % game type to solve:
    % "max": neighbors cooperative, "min": neighbors opponents
CFG.rob = 0;                    % 1: consider only neighbors on robust graph; 0: all neighbors
CFG.criterion = 'optim';        % expansion criterion: optim(istic) or unif(orm)
CFG.iagent = [];                % index of this agent
CFG.G = [];                     % graph of interagent connections: adjacency matrix
CFG.Grob = [];                  % minimal robust graph (required if rob=1)
CFG.U = [];                     % discrete, flat action spaces of ALL agents, in a cell array
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

maxmin_fun = eval(['@' cfg.gametype]);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; optimistic = strcmp(cfg.criterion, 'optim');
iagent = cfg.iagent; 
if cfg.rob,
    g = cfg.Gstar(iagent, :);
    if ~all(cfg.G(iagent, g)),
        message = sprintf('WARNING! Robust graph connection of [%d] with [%d] broken, cannot run planning', ...
            iagent, find(g & ~cfg.G(iagent, :)));
        dispx(message, cfg.verb, 0);
        uroot = NaN;
        planinfo.n = []; planinfo.depth = []; planinfo.trun = []; planinfo.treesize = [];
        planinfo.message = message; 
        plandata = [];
        return;
    end;
else
    g = cfg.G(iagent, :);
    if ~any(g), 
        message = sprintf('WARNING! All connections of [%d] broken, cannot run planning', iagent); 
        dispx(message, cfg.verb, 0);
        uroot = NaN;
        planinfo.n = []; planinfo.depth = []; planinfo.trun = []; planinfo.treesize = [];
        planinfo.message = message; 
        plandata = [];
        return;
    end;
end;
iothers = find(g);
pranges =  cfg.pranges;

% form state value
x = xroot(pranges(iagent)+1:pranges(iagent+1));
for iother = iothers, x = [x; xroot(pranges(iother)+1:pranges(iother+1))]; end;
xroot = x;
dimx = sum(model.p([iagent iothers])); % number of state vars (computed differently as verification)
% form action spaces
Uagent = flat(cfg.U{iagent});
Magent = size(Uagent, 2);           % number of discrete actions for agent
Uothers = flat({cfg.U{iothers}});   % actions of others               
Mothers = size(Uothers, 2);         % same, for the others    

% start counting time
t = cputime;

maxtreesize = Magent*Mothers*n + 1;
% prepare tree data 
Tparent = zeros(3, maxtreesize, 'uint32');        % parent and parent action: for agent and others
Tchildren = zeros(Magent, Mothers, maxtreesize, 'uint32');  % child index for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tr = NaN(1, maxtreesize, 'double');               % reward
Tterm = false(1, maxtreesize);                    % terminal flag
Tb = NaN(1, maxtreesize, 'double');               % b-values (upper bounds)
Tutil = NaN(1, maxtreesize, 'double');            % nu-values (lower bounds)

% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(:, root) = 0;
Tchildren(:, :, root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:, root) = xroot;
Tr(root) = 0;
Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state
Tb(root) = 1/(1-discount);
Tutil(root) = 0;

treesize = 1;

% 2) Main tree construction loop
for tau = 1:n,
    
    % 2a) select candidate leaves for expansion
    if optimistic,      % optimistic: navigate down maxmin or maxmax path
        node = root;
        while ~Tleaf(node), 
            % note ties are always broken for the first min/maximizer for others; first maximizer
            % for self
            B = Tr(Tchildren(:, :, node)) + discount .* Tb(Tchildren(:, :, node));
            [B, jothers] = maxmin_fun(B, [], 2); [B, jagent] = max(B);
            node = Tchildren(jagent, jothers(jagent), node);
        end;
        % Note a limitation: if the leaf found is terminal nothing happens, even
        % if by breaking ties we could reach another leaf
        if Tterm(node), dispx('WARNING! Optimistic leaf is terminal. Stopping algorithm.', cfg.verb, 0); break; end;
        nodetoexpand = node;
    else                % assumed uniform: smallest depth
        leaves = find(Tleaf);
        leaves = leaves(Tdepth(leaves) == min(Tdepth(leaves)));
        % select only nonterminal leaves; if none, nothing left to do
        leaves = leaves(~Tterm(leaves)); if isempty(leaves); break ;end;
        % break any ties between remaining candidates deterministically
        nodetoexpand = leaves(1);
    end;
        
    % 2b) expand the chosen node
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions 
    for jagent = 1:Magent,
        for jothers = 1:Mothers,
            % find index where new node will be, save it on the children table of current node
            child = treesize + 1;
            Tchildren(jagent, jothers, nodetoexpand) = child;

            % simulate transition and add data to new node
            [Tx(:, child), Tr(child), Tterm(child)] = feval(model.fun, model, Tx(:, nodetoexpand), ...
                [Uagent(:, jagent); Uothers(:, jothers)], 'pred', iagent, g);
            if Tr(child) < 0 || Tr(child) > 1, error('Rewards must be in [0, 1]'); end;

            % update pointers and other data for new node
            Tparent(:, child) = [nodetoexpand; jagent; jothers];      
            % parent is expanded node, along jagent x jothers action
            Tchildren(:, :, child) = 0;                 % init children pointers to "null"
            Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
            Tleaf(child) = true;                        % new child is a leaf
            % upper and lower bounds
            if Tterm(child),    Tb(child) = 0;          % values of terminal nodes exactly 0
            else                Tb(child) = 1/(1-discount);
            end;
            Tutil(child) = 0;

            % we just added a new node, increase tree size
            treesize = treesize + 1;
        end;
    end;
    
    % 2c) update b- and nu-values affected by expanded node (path to root)
    node = nodetoexpand;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        children = Tchildren(:, :, node);
        Tb(node) = max(maxmin_fun(Tr(children) + discount .* Tb(children), [], 2));
        Tutil(node) = max(maxmin_fun(Tr(children) + discount .* Tutil(children), [], 2));
        node = Tparent(1, node);
    end;

end;    % end main tree construction loop

% 4) Find action at the tree root
B = Tr(Tchildren(:, :, root)) + discount .* Tb(Tchildren(:, :, root));
[B, jagent] = max(maxmin_fun(B, [], 2));
uroot = Uagent(:, jagent);

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
end;
% output only tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, :, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
    plandata.Tb = Tb(1:treesize);
    plandata.Tutil = Tutil(1:treesize);
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions



