function [xplus, rplus, terminal, G] = mahamm_mg(m, x, u, mode, iagent, G, xplusothers)
%  Discrete-time dynamics of the Markov game for decoupled, homogeneous linear agents.
%  [XPLUS, RPLUS, TERMINAL, G] = MAHAMM_MG(M, X, U, MODE, IAGENT, G)
%  Parameters:
%   M           - the model specification
%   X           - current state, x(k), as a flat vector
%   U           - commands u(k) as a flat vector
%       Both X and U should contain states for all agents (in 'real' mode) 
%   MODE        - 'real' for "real" complete transition (only supported value)
%   IAGENT      - unused
%   G           - in mode 'pred', connections of this agent, as an adjacency row vector
%   XPLUSOTHERS - unused
%  Returns:
%   XPLUS       - agent states at next sample, x(k+1)
%   RPLUS       - ensuing rewards, r(k+1)
%   TERMINAL    - whether the state is terminal (i.e., end of episode)
%               Should always be 0 for a continuing task
%   G           - graph of all agents (in 'real' mode)

if mode(1) ~= 'r',  error('Only real mode supported'); end;
nagents = m.nagents;

% simulate the agents
xa = reshape(x, m.p(1), nagents);
ua = reshape(u, m.q(1), nagents);
xplus = nan(size(xa));
for i = 1:nagents,
    xplus(:, i) = m.A * xa(:, i) + m.B * m.inputnl(ua(:, i));
end;

if m.rewtype(1) == 'n',        % 'none'
    rplus = nan(nagents, 1);
    G = m.G;
end;

% flatten state
xplus = xplus(:);

% task is non-episodic
terminal = 0;

% END function
end