function [xplus, rplus, terminal, G] = malinear_mg(m, x, u, mode, iagent, G, xplusothers)
%  Discrete-time dynamics of the Markov game for decoupled, homogeneous linear agents.
%  [XPLUS, RPLUS, TERMINAL, G] = MALINEAR_MG(M, X, U, MODE, IAGENT, G)
%  Parameters:
%   M           - the model specification
%   X           - current state, x(k), as a flat vector
%   U           - commands u(k) as a flat vector
%       Both X and U should contain states for all agents (in 'real' mode) and for the current
%       agents and those connected to it (in 'pred' mode), see below
%   MODE        - 'real' for "real" complete transition (default value)
%                 'pred' for prediction from the perpective of a single agent; state and action should be
%                 arranged so that the first is of this agent, and then G
%   IAGENT      - in mode 'pred', index of this agent
%   G           - in mode 'pred', connections of this agent, as an adjacency row vector
%   XPLUSOTHERS - in mode 'pred', the next states of the other agents (to avoid simulating); 
%               if only some agent states are available, the others should be NaN
%               Note since the system is very simple, using this option will likely not pay off
%               Flat vector containing only the states of the neighbors
%  Returns:
%   XPLUS       - agent states at next sample, x(k+1)
%   RPLUS       - ensuing rewards, r(k+1)
%   TERMINAL    - whether the state is terminal (i.e., end of episode)
%               Should always be 0 for a continuing task
%   G           - graph of all agents (in 'real' mode) or connections of this agent ('pred' mode); 
%               in prediction mode, range checks on the agents are not performed and G remains
%               unchanged

if mode(1) == 'r',  nagents = m.nagents;
else                nagents = 1 + sum(G);
end;
if nargin < 7,      xplusothers = []; end;

if isempty(xplusothers),
    % simulate everything
    xplus = m.A * reshape(x, m.p(1), nagents) + m.B * reshape(u, m.q(1), nagents);
else
    x = reshape(x, m.p(1), nagents); u = reshape(u, m.q(1), nagents);
    xplusothers = reshape(xplusothers, m.p(1), nagents - 1);
    % simulate current agent
    xplus(:, 1) = m.A * x(:, 1) + m.B * u(:, 1);
    % for the rest: if state not available, agent must be simulated,
    % otherwise, use the available state
    for i = 2:nagents,
        if any(isnan(xplusothers(:, i-1))), xplus(:, i) = m.A * x(:, i) + m.B * u(:, i);
        else                                xplus(:, i) = xplusothers(:, i-1);
        end;
    end;
end;

% note 'flock' reward is taken equivalent to 'varconn' graph, and 'none' reward with 'fixedconn'
% this should be fine given the error checks in malinear_problem
if m.rewtype(1) == 'f',     % flock, we can assume double int
    % xplus = reshape(x, m.p(1), nagents); % for debugging purposes
    pos = xplus(1:m.dim, :);
    vel = xplus(m.dim+1:end, :);
    if mode(1) == 'r',
        % compute distances between each pair of agents
        dp = repmat(pos(:), 1, nagents) - repmat(pos, nagents, 1);
        dp = squeeze(sqrt(sum(reshape(dp, m.dim, nagents, nagents).^2, 1)));
        % check who the neihbors are (closer than comm range)
        G = dp <= m.range;
        G(1:nagents+1:nagents*nagents) = 0;     % set diagonal to 0 (no neighb to itself)

        % compute differences in velocities between all pairs of agents
        dvsq = repmat(vel(:), 1, nagents) - repmat(vel, nagents, 1);
        dvsq = squeeze(sum(reshape(dvsq, m.dim, nagents, nagents) .^ 2, 1));

        % reward: error distance-setpoint, velocity diff, summed over G
        % note that because range < flock_maxdpos => neighb=0, we don't need to cap distances
        if m.flock_dist < m.range,
            dp = max(dp, m.flock_dist);
            ppenalty = sum(G.*(dp - m.flock_dist).^2, 2) ./ m.flock_pnorm;
        else    % step function
            ppenalty = sum(G.*(dp > m.range), 2);
        end;
        vpenalty = sum(min(G.*dvsq, m.flock_vnorm), 2) ./ m.flock_vnorm;
        % reward per agent
        rplus = - m.flock_alpha * ppenalty - m.flock_beta * vpenalty;
        % deal explicitly with agents without any connection, to avoid division by 0
        connected = any(G, 2);
        % normalize
        rplus(connected) = 1 + (rplus(connected) ./ sum(G(connected, :), 2));
        rplus(~connected) = 0;
    else
        % prediction mode, similar but now:
        % - just the perspective of the first agent matters
        % - agents do not drop out of comm range (since this is just
        % prediction not real interaction)
        dp = repmat(pos(:, 1), 1, nagents) - pos; dp = sqrt(sum(dp.^2, 1));
        dvsq = repmat(vel(:, 1), 1, nagents) - vel; dvsq = sum(dvsq.^2, 1);
        % reward (note missing 2:end from dvsq, it's irrelevant since dvel(1) is always 0)
        % normalized penalty for position velocity differences
        if m.flock_dist < m.range,
            dp = min(max(dp, m.flock_dist), m.range);
            ppenalty = sum((dp(2:end) - m.flock_dist).^2) / m.flock_pnorm;
        else    % step function
            ppenalty = sum(dp(2:end) > m.range);
        end;
        vpenalty = sum(min(dvsq, m.flock_vnorm)) / m.flock_vnorm;
        % overall normalized reward
        rplus = 1 - (m.flock_alpha * ppenalty + m.flock_beta * vpenalty) / (nagents-1);
    end;
elseif m.rewtype(1) == 'n',        % 'none'
    if mode(1) == 'r', 
        rplus = nan(nagents, 1);
    else
        rplus = nan;
    end;
    G = m.G;
end;

% flatten state
xplus = xplus(:);

% task is non-episodic
terminal = 0;

% END function
end