function [xplus, rplus, terminal, G] = manonhol_mg(m, x, u, mode, iagent, G, xplusneighb)
%  Discrete-time dynamics of the Markov game for decoupled nonholonomic agents.
%  [XPLUS, RPLUS, TERMINAL, G] = MANONHOL_MG(M, X, U, MODE, IAGENT, G)
%  Parameters:
%   M           - the model specification
%   X           - current state, x(k), as a flat vector
%   U           - commands u(k) as a flat vector
%       Both X and U should contain states for all agents (in 'real' mode) and for the current
%       agents and those connected to it (in 'pred' mode), see below
%   MODE        - 'real' for "real" complete transition (default value)
%                 'pred' for prediction from the perpective of a single agent; state and action should be
%                 arranged so that the first is of this agent, and then the neighbors in G
%   IAGENT      - in mode 'pred', index of this agent
%   G           - in mode 'pred', connections of this agent, as an adjacency row vector
%   XPLUSNEIGHB - in mode 'pred', the next states of the other agents (to avoid simulating); 
%               if only some agent states are available, the neighb should be NaN
%               Flat vector containing only the states of the neighbors
%  Returns:
%   XPLUS       - agent states at next sample, x(k+1)
%   RPLUS       - ensuing rewards, r(k+1) (zeros in 'real' mode)
%   TERMINAL    - whether the state is terminal (i.e., end of episode)
%               Should always be 0, this is a continuing task
%   G           - graph of all agents (in 'real' mode) or connections of this agent ('pred' mode); 
%               in prediction mode, range checks on the agents are not performed and G remains
%               unchanged

if mode(1) == 'r',  
    % real transition -- just simulate everyone, rewards are not interesting
    nagents = m.nagents; 
    x = reshape(x, m.p(1), nagents); u = reshape(u, m.q(1), nagents);
    xplus = zeros(size(x));
    for i = 1:nagents,
        xplus(:, i) = manonhol_dttrans(m, x(:, i), u(:, i));
    end;
    rplus = zeros(1, m.nagents);
    if m.matype(1) == 'v',
        G = manonhol_buildgraph(m, xplus);
    else 
        G = m.G;
    end;
else
    % predicted transition
    if nargin < 7,  xplusneighb = []; end;
    
    nagents = 1 + sum(G);
    x = reshape(x, m.p(1), nagents); u = reshape(u, m.q(1), nagents);
    % find the predicted states of the other agents if supplied
    if ~isempty(xplusneighb),
        xplusneighb = reshape(xplusneighb, m.p(1), nagents - 1); 
    end;
    % simulate current agent
    xplus(:, 1) = manonhol_dttrans(m, x(:, 1), u(:, 1));
    % for the rest: if state available, use it, else simulate it
    for i = 2:nagents,
        if ~isempty(xplusneighb) && ~any(isnan(xplusneighb(:, i-1))), 
            xplus(:, i) = xplusneighb(:, i-1);
        else
            % should never happen with the exper settings in this case
            warning('simulating a neighbor');
            xplus(:, i) = manonhol_dttrans(m, x(:, i), u(:, i));
        end;
    end;
    
    % compute reward
    xa = xplus(m.agree, :);
    xc = xplus(m.conn, :);
    costagree = 0; rewconn = 0;
    for i = 2:nagents,
        % agreement cost sums up the weighted-norm distances, capped to 1
        costagree = costagree + min(1, sqrt((xa(:, 1) - xa(:, i))' * m.Q * (xa(:, 1) - xa(:, i))));
        % connectivity reward if the agents are connected
        rewconn = rewconn + (sqrt(sum((xc(:, 1) - xc(:, i)).^2)) <= m.range);
    end;
    rplus = (1-m.beta) * (1 - costagree/(m.nagents - 1)) + m.beta * rewconn / (m.nagents - 1);
        if isfield(m, 'brstop') && m.brstop,
        end;
end;

% flatten state
xplus = xplus(:);
% task is non-episodic
terminal = 0;

% END function
end

% discrete-time transition function -- for now just Euler supported
function xplus = manonhol_dttrans(m, x, u)
    xplus = x + m.Ts .* manonhol_cttrans(0, x, m, u);
end
% continuous-time transition function
function dx = manonhol_cttrans(t, x, m, u)
    v = x(3);
    theta = x(4);
    dx = [v*cos(theta); v*sin(theta); u(1); u(2)];
end

