function varargout = mararm_problem(what, varargin)
% Multiagent robotic arm problem: decoupled dynamics, two-link robot arms.
%   OUT = RARM_PROBLEM(MODE, CONFIG)
%   [X0, G] = RARM_PROBLEM('reset', CFG, RESETCONFIG);

switch what,
    % Offer basic info about the model (without creating the actual model)
    case 'shortid', varargout = {'mararm'};
    case 'info',    
        info.id = 'mararm';               % an identifier
        info.problem = @mararm_problem;   % handle to model function
        info.det = 1;                     % whether problem is deterministic or stochastic
        varargout = {info};
    
    % Create model
    case 'model',        
        % default params
        % 1) multirobot system parameters
        CFG.nagents = [];                       % number of agents (robots)
        CFG.G = [];                             % communication graph
        
        % 2) parameters of each robot arm (MAS assumed homogeneous, all robot arms identical)
        % physics
        CFG.m =        [1.25    .8];           % [kg] link masses
        CFG.len =      [.4     .4];            % [m] link lenghts
        CFG.c =        CFG.len ./ 2;           % [m] coordinates of centers of mass along links
        CFG.b =        [0.08     0.02];        % [kg/sec] joint dampings 
        CFG.I =        CFG.m .* (CFG.len.^2)/3;% [kg.m^2] inertias, proportional with mass 
        CFG.maxomega = [2*pi;  2*pi];          % [rad/sec] maximum angular speeds of links
        CFG.maxtau =   [1.5;    1];            % [Nm] maximum torques
        CFG.orientation = 'horiz';             % 'horiz' or 'vert', arm orientation
        % rewards / goal
        CFG.rewtype = 'quadcons';              % (normalized) quadratic reward for consensus 
        CFG.Q = diag([1 1 1 1]);               % state weights for quadratic consensus rewards
        % numerical integration
        CFG.Ts = .05; CFG.odemethod = 'fixed-ode'; CFG.odesolver = 'ode4'; CFG.odesteps = 5;
        
        % process any user-configured fields
        model = parseconfig(varargin{1}, CFG);
        
        % create standard model variables (these are "set in stone") 
        % suffix "1" means the variables are for a single agent
        model.id = 'mararm';                    % an identifier
        model.matype = 'fixedconn';             % connection graph does not vary
        model.p1 = 4; model.q1 = 2;             % always 4 states, 2 actions per robot
        model.p = 4 + zeros(1, model.nagents); model.q = 2 + zeros(1, model.nagents); 
        model.pranges = [0 cumsum(model.p)];  model.qranges = [0 cumsum(model.q)];
        % set bounds in standard format
        model.maxx1 = [pi; model.maxomega(1); pi; model.maxomega(2)]; 
        model.maxu1 = model.maxtau(:); 
        model.maxx = repmat(model.maxx1, model.nagents, 1);
        model.maxu = repmat(model.maxu1, model.nagents, 1);  
        model.simagent = @mararm_simagent;
        model.fun = @mararm_mg;
        model.plotfun = @mararm_plot;
        
        % compute normalization coefficient for rewards
        switch model.rewtype,
            case 'quadcons',    % LQR normalized to interval [0, 1]
                % note the maximal angular difference is pi, since distance wraps around the circle
                maxdiff = [pi; 2*model.maxomega(1); pi; 2*model.maxomega(2)];
                model.nonnorm_maxr = maxdiff' * model.Q * maxdiff;
                model.maxr = 1;
            case 'none',
        end;
        
        % compute physical helper vars
        model.g = 9.81;
        model.P1 = model.m(1) * model.c(1)^2 + model.m(2) * model.len(1)^2 + model.I(1);
        model.P2 = model.m(2) * model.c(2)^2 + model.I(2);
        model.P3 = model.m(2) * model.len(1) * model.c(2);
        model.g1 = (model.m(1) * model.c(1) + model.m(2) * model.len(1)) * model.g;
        model.g2 = model.m(2) * model.c(2) * model.g;

        % create discretization helper vars
        switch model.odemethod,
            case 'ode',         model.odet = [0 model.Ts/2 model.Ts]; model.odeopt = odeset;
            case 'fixed-ode',   model.odet = 0 : model.Ts / model.odesteps : model.Ts;
            case 'euler',       % do nothing
        end;
        
        varargout = {model};
        
    case 'reset',   % handle state reset
        RES.type = [];
        RES.x0 = [];
        m = varargin{1};
        res = varargin{2};
        if ischar(res) && any(strcmp(res, {'rand', 'randangles', 'values'})),   % handle simple method identifier
            res = struct('type', res);
        elseif isnumeric(res),
            res = struct('type', 'values', 'x0', res);
        end;
        
        res = parseconfig(res, RES);
        switch res.type,
            case 'rand',    x0 = (2*rand(sum(m.p), 1) - 1) .* m.maxx;
            case 'randangles', 
                x0 = [(2*rand(1, m.nagents) - 1) .* pi;
                    zeros(1, m.nagents); 
                    (2*rand(1, m.nagents) - 1) .* pi; 
                    zeros(1, m.nagents)];
            case 'values',  x0 = res.x0;
        end;          
        % just return the fixed connection graph
        G = m.G;
    
        % flatten and output x, output G
        varargout = {x0(:), G};

end;

% END ====================================================================

