function varargout = lspih(cfg)
% Least-squares policy iteration with explicitly parameterized policies.
%   [THETA, HTHETA] = LSPIH(CFG)     - in 'run' mode
%   [HIST, FIGH] = LSPIH(CFG)        - in 'replay' mode
%   FIGH = LSPIH(CFG)                - in 'sol', 'evol', 'examinetheta' modes
% An implementation of least-squares policy iteration [1]. In contrast to the original 
% algorithm, this version uses explicitly parameterized policies, and supports constrained
% policy improvements.
%
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
%
% Outputs:
%       THETA       - the computed Q-function parameter vector
%       HTHETA      - the computed policy parameter vector
%       HIST        - the replay history (trajectories)
%       FIGH        - handles to the figures created
%
% [1] Lagoudakis, M. G. & Parr, R. "Least-Squares Policy Iteration"
%   Journal of Machine Learning Research, 2003, 4, 1107-1149

% Author: Lucian Busoniu
% Version: 2.0
% History:
%   1.0     - initial version
%   1.0     - added test for oscillations, in addition to convergence

% Detailed execution time statistics are stored in variable "timestat". Fields:
%   qsamples        - scalar, time to generate the policy eval samples
%   hsamples        - scalar, time to generate the policy improvement samples
%   precomputephi   - scalar, time to precompute Q BF values (if precomputed)
%   precomputhphi   - scalar, time to precompute h BF values (if precomputed)
%   precomputeAb    - scalar, time to precompute fixed part of A, and b (if precomputed)
%   init            - scalar, aggregate initialization time (sum of all of the above)
%   iter_computeAb  - vector, time/iter to compute A and b 
%   iter_qsolve     - vector, time/iter to solve A theta = b
%   iter_computeh   - vector, time/iter to compute h samples
%   iter_hsolve     - vector, time/iter to solve policy (constrained) LSQR problem
%   iter_peval      - vector, time/iter to perform policy eval (iter_computeAb + qsolve)
%   iter_pimpr      - vector, time/iter to perform policy improvement (iter_computeh + hsolve)
%   iter_convtest   - vector, time/iter to test convergence and oscillation
%   iter_run        - vector, total runtime/iter (iter_peval + iter_pimpr + iter_convtest)
%   run             - scalar, aggregate runtime (sum of all iter_run)
% For compatibility reasons, a variable called "trun" is also kept, equal
% to timestat.run

% TODO implement index optimization; pre-computation of b and A0

if nargin < 1, cfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
% function config
CFG.run = 0;                        % run learning
CFG.resume = 0;                     % resume learning
CFG.replay = 0;                     % replay learned policy
CFG.sol = 0;                        % plot solution and convergence stats
CFG.evol = 0;                       % plot evolution statistics
CFG.exth = [];                      % configuration to examine theta (leave empty not to examine)
CFG.datafile = 'lspidata';          % save data to file
CFG.datadir = [];                   % data dir
% main algorithm config
CFG.problem = '';                   % what problem to solve
CFG.approx = [];                    % Q-function approximator object/config
CFG.samples = [];                   % collection of samples object/config for LSTD-Q
CFG.happrox = [];                   % policy approximator
CFG.hsamples = [];                  % collection of samples for policy improvement
CFG.con = [];                       % policy constraint object/config (empty if none)
CFG.gamma = 0.95;                   % discount factor
CFG.maxiter = 100;                  % maximum # of iterations
CFG.invdelta = 0;                   % use for inverse computation, ridge regression
CFG.loadsamples = [];               % load LSTD-Q samples from this file
CFG.loadhsamples = [];              % load policy improvement samples from this file
CFG.h0 = [];                        % start with this initial policy
CFG.h0args = {};                    % arguments for the initial policy
% convergence and oscillation checks
CFG.eps = 0.01;                     % convergence threshold
CFG.checkq = 1;                     % also verify conditions for the Q parameter
CFG.maxper = 10;                    % osc: maximum period to check (# iter)
CFG.nper = 1;                       % osc: number of periods to check
CFG.epsper = 0.01;                  % osc: maximum abs diff between two par vectors in diff periods
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% stats config
CFG.storeab = 0;                    % whether to store A and b histories
% display config
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.iterdisp = 1;                   % feedback after every iterdisp iterations
CFG.itersave = 3;                   % save after each itersave iterations
CFG.sampledisp = .05;               % display progress after this % of samples processed
CFG.silent = 0;                     % suppress all output
% figure config
CFG = setfigprop(CFG, 'addfields');
CFG.miscinfo = '';                  % field for miscellaneous information

% Early defaults (initialized before calling problem defaults)
ECFG.model_params = {};             % parameters for problem calling in 'model' mode
ECFG.lspi_params = {};              % parameters for problem calling in 'lspi' mode

% List of fields that define the problem
KEEPFIELDS = {'problem', 'gamma', 'approx', 'samples', 'happrox', 'hsamples', 'usesparse'};

% ==== PARSE AND PROCESS CONFIG ====
cfg = parseconfig(cfg, CFG, ECFG, 'lspi');
% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       
% determine if initialization needed; if not, need to load: check if datafile exists
cfg.init = ~(cfg.resume || cfg.replay || cfg.sol || cfg.evol || ~isempty(cfg.exth));
if ~cfg.init && ~exist([cfg.datafile '.mat'], 'file'),
    error(['File [' cfg.datafile '.mat] does not exist.']);
end;

% ==== IF NOT INITIALIZING: NEED TO LOAD DATA ====
if ~cfg.init,        % load data file, making sure that cfg and KEEPFIELDS is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
    cfg.approx = revise_approx(cfg.approx);     % make sure approx is in latest format
    approx = revise_approx(approx);    % for replay, etc functions that assume approx is initialized
    cfg.samples = revise_samples(cfg.samples);  % the same for samples
    cfg.happrox = revise_approx(cfg.happrox);
    happrox = revise_approx(happrox); 
end;

% Echo config
dispx('LSPI will run with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');


% ==== IF INITIALIZING: CREATE MODEL, AND IF NEEDED: APPROXIMATOR, SAMPLES ====
if cfg.init,
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    % heuristic check if we already have a Q approx object
    if isstruct(cfg.approx) && isfield(cfg.approx, 'N') && isfield(cfg.approx, 'phi') ...
            && isfield(cfg.approx, 'q') && isfield(cfg.approx, 'h'),
            % nothing to do
            dispx('Q-function approx object supplied', cfg.verb, 2);
    else    % create approximator using cfg.approx as the config
        cfg.approx = create_approx(model, cfg.approx);
    end;
    % LSTD-Q samples object
    if ~isempty(cfg.loadsamples) && exist([cfg.loadsamples '.mat'], 'file'),
        % try loading samples -- performs no checking whether the same type & number of
        % samples is required
        loadwithprefix(cfg.loadsamples, 's_', 'cfg');
        cfg.samples = s_cfg.samples;
        clear s_cfg;
        dispx(['LSTD-Q samples object loaded from [' cfg.loadsamples ']'], cfg.verb, 1);
    else
        % heuristic check if we already have a samples collection supplied
        if isstruct(cfg.samples) && isfield(cfg.samples, 'N') && isfield(cfg.samples, 'X') ....
                && isfield(cfg.samples, 'U') && isfield(cfg.samples, 'Xp') ...
                && isfield(cfg.samples, 'R'),
                dispx('LSTD-Q samples object supplied', cfg.verb, 1);
            % nothing to do
        else
            % generate samples using cfg.samples as the config
            % count this in the init time
            dispx('Generating LSTD-Q samples...', cfg.verb, 0);
            tmark_init = cputime;
            cfg.samples = generate_samples(model, cfg.samples, cfg.approx);
            timestat.qsamples = cputime - tmark_init; clear tmark_init;
            timestat.init = timestat.qsamples;
            dispx([8 ' done.'], cfg.verb, 0);
        end;
    end;
    % heuristic check if we already have a policy approx object
    if isstruct(cfg.happrox) && isfield(cfg.happrox, 'N') && isfield(cfg.happrox, 'phi') ...
            && isfield(cfg.happrox, 'h'),
            % nothing to do
            dispx('Policy approx object supplied', cfg.verb, 1);
    else    % create approximator using cfg.approx as the config
        cfg.happrox = create_approx(model, cfg.happrox);
    end;
    % policy improvement samples object
    if ~isempty(cfg.loadhsamples) && exist([cfg.loadhsamples '.mat'], 'file'),
        % try loading samples -- performs no checking whether the same type & number of
        % samples is required
        loadwithprefix(cfg.loadsamples, 's_', 'cfg');
        cfg.hsamples = s_cfg.hsamples;
        clear s_cfg;
        dispx(['Policy improvement samples object loaded from [' cfg.loadhsamples ']'], cfg.verb, 1);
    else
        % heuristic check if we already have a samples collection supplied
        if isstruct(cfg.hsamples) && isfield(cfg.hsamples, 'N') && isfield(cfg.hsamples, 'X'),
            dispx('Policy improvement samples object supplied', cfg.verb, 1);
            % nothing to do
        else
            % generate samples using cfg.samples as the config
            % count this in the init time
            dispx('Generating policy improvement samples...', cfg.verb, 0);
            tmark_init = cputime; 
            cfg.hsamples = generate_xsamples(model, cfg.hsamples, cfg.happrox);
            timestat.hsamples = cputime - tmark_init; clear tmark_init;
            timestat.init = timestat.init + timestat.hsamples;
            dispx([8 ' done.'], cfg.verb, 0);
        end;
    end;
    % policy constraints
    if ~isempty(cfg.con),
        dispx('Creating constraint object...', cfg.verb, 0);
        cfg.con = create_constraint(model, cfg.approx, cfg.con);
        dispx([8 ' done.'], cfg.verb, 0);
    end;
    % initial policy
    if ~isempty(cfg.h0),
        % determine the order of the possibly dynamic controller
        try     [cx0, cord] = feval(cfg.h0, 'init', cfg.h0args{:});
        catch   cord = 0;   % compatibility for old-style static policies
        end;
        if cord > 0, error('Cannot control non-trajectory samples with a dynamical controller'); end;    
    end;
end;

% cfg.approx
% cfg.samples

% ==== RUN LSPI ====
if cfg.run || cfg.resume,
       
    % shorthand variables
    approx = cfg.approx;
    N = approx.N;
    Ns = cfg.samples.N;
    Xs = cfg.samples.X; Us = cfg.samples.U; Xps = cfg.samples.Xp; Rs = cfg.samples.R;

    happrox = cfg.happrox;
    hN = happrox.N;
    hNs = cfg.hsamples.N;
    hXs = cfg.hsamples.X;
    
    con = cfg.con;

    % Initialize if not resuming
    if cfg.init,
        % History variables
        thetah = cell(cfg.maxiter+1, 1);
        deltah = nan(cfg.maxiter+1, 1);     % deltah(1) will always remain NaN
        hthetah = cell(cfg.maxiter+1, 1);
        hdeltah = nan(cfg.maxiter+1, 1);    % hdeltah(1) will always remain NaN
        condA = nan(cfg.maxiter, 1);        % condition number history for A
        if cfg.storeab, Ah = cell(cfg.maxiter, 1); bh = cell(cfg.maxiter, 1); end;

        % if worst-case storage space for pre-computed BF values of all the (x, u) samples
        % is reasonable, pre-init. For now, fixed to 1 (preinit) for efficiency
        storephi = 1; %(N * Ns * 8) <= 300e6;   % 300 MB with double (8 bytes) storage
        
        % init constraints
        if ~isempty(con),
            if con.linear,
                % get constraint matrix and vector
                [Acon, bcon] = happrox.lincon(happrox, con);
                % init lsqrlin options
                % when we use constraints, we can't use Large Scale algorithm anyway
                % MTB 2015 complains that interior point does not work with
                % constraints so we set to active-set explicitly in that
                % version, and also from R2010a onwards (where Algorithm is
                % supported)
                if verLessThan('Matlab', '7.10'),
                    optcon = optimset('MaxIter', 500, 'LargeScale', 'off', 'Display', 'off');
                else
                    optcon = optimset('MaxIter', 500, 'Algorithm', 'active-set', 'LargeScale', 'off', 'Display', 'off');
                end;
            else
                error('LSPIH currently only supports linear constraints.');
            end;
        end;
        
        % init helpers for discrete-action Q-approximator
        disc = any(strcmp(approx.type, {'rbfdisc', 'triang'}));
        if disc,        % action grids and their size
            U = approx.U; 
            Un = zeros(length(U)); for i=1:length(U), Un(i) = length(U{i}); end;        
        end;
        
        % init running vars
        k = 1;                  % iteration index
        timestat.run = 0;       % execution time
        % initialize parameter vectors to zeros (& save them on the histories)
        theta = zeros(N, 1);
        htheta = zeros(hN, 1);
        thetah{1} = theta; 
        hthetah{1} = htheta;
    end;        % initialization IF
    
    % use sparse A and PHI if the approx type recommends it
    if isfield(approx, 'sparse') && approx.sparse,  cfg.usesparse = 1; zeromatrix = @sparse;
    else                                            cfg.usesparse = 0; zeromatrix = @zeros;
    end;
    if ~storephi,
        dispx('WARNING! Cannot store LSTD-Q BF values. Algorithm will run slowly.', cfg.verb, 0);
    end;
    if storephi && (cfg.init || ~exist('PHI', 'var')),
        if cfg.init,    
            dispx('Pre-computing BF values for LSTD-Q samples...', cfg.verb, 0);            
            tmark_init = cputime;
        else
            dispx('Re-computing BF values for LSTD-Q samples...', cfg.verb, 0);
        end;
        PHI = zeromatrix(N, Ns);
        for i = 1:Ns, PHI(:, i) = approx.phi(approx, Xs(:, i), Us(:, i)); end;
        % Only the first computation of BF values is counted in
        % the init time (otherwise tinit would grow with each recomputation)
        if cfg.init, 
            timestat.precomputephi = cputime - tmark_init; clear tmark_init;
            timestat.init = timestat.init + timestat.precomputephi; 
        end;
        dispx([8 ' done.'], cfg.verb, 0);
    end;

    % (Re)compute BF values for the policy improvement samples. Note BF vectors are on rows, 
    % so that the matrix is ready for the LSQR problem without transposition (for speed)
    % (This matrix is always necessary in the crt implementation)
    if cfg.init || ~exist('HPHI', 'var'),
        % use sparse HPHI the policy approx type recommends it
        if isfield(happrox, 'sparse') && happrox.sparse,    cfg.husesparse = 1; HPHI = sparse(hNs, hN);
        else                                                cfg.husesparse = 0; HPHI = zeros(hNs, hN);
        end;
        if cfg.init,    
            dispx('Pre-computing BF values for h-improve samples...', cfg.verb, 0);
            tmark_init = cputime;
        else
            dispx('Re-computing BF values for h-improve samples...', cfg.verb, 0);
        end;
        for i = 1:hNs, HPHI(i, :) = happrox.phi(happrox, hXs(:, i)); end;
        % note only the first computation is counted in the init time
        if cfg.init, 
            timestat.precomputehphi = cputime - tmark_init; clear tmark_init;
            timestat.init = timestat.init + timestat.precomputehphi; 
        end;
        if cfg.husesparse, condHPHI = cond(full(HPHI));  % dirty shortcut, needs to be fixed...
        else condHPHI = cond(HPHI);
        end;
        dispx([8 ' done.'], cfg.verb, 0);
    end;
    
    dispx('Performing LSPI...', cfg.verb, 0);
    
    % check if should use an initial policy
    useh0 = ~isempty(cfg.h0);
    % Main policy iteration loop
    conv = 0; Anoninv = 0; HPHIpoor = 0; 
    while ~conv && ~Anoninv && ~HPHIpoor && k <= cfg.maxiter,
        tmark = cputime;
        
        % Perform policy evaluation, LSTD-Q
        % 0) init A and b
        if cfg.invdelta > 0,
            if cfg.usesparse,   A = sparse(1:N, 1:N, cfg.invdelta, N, N);
            else                A = eye(N, N) .* cfg.invdelta;
            end;
        else                    A = zeromatrix(N, N);
        end;
        b = zeros(N, 1);        % b needs not be sparse
        % a) loop through samples, updating A and b
        progs = cfg.sampledisp; dispx(sprintf('k=%d: LSTD-Q, samples processed: %2d%%', k, 0), cfg.verb, 3);
        for i = 1:Ns,
            % find phi(x, u) (note we should actually check storePHI here)
            phixu = PHI(:, i);
            % find out the action in x'
            if k == 1 && useh0,         % use an initial policy
                up = cfg.h0(Xps(:, i), [], cfg.h0args{:});  % [] = undef time arg
            else                        % use the current poolicy
                up = happrox.h(happrox, htheta, Xps(:, i));
            end;
            % quantize action given by policy if using a discrete-action Q-approximator
            if disc, up = quantize(up, U, Un); end; 
            % compute phi(x', u')
            phixpup = approx.phi(approx, Xps(:, i), up);
            % update A and b
            A = A + phixu * (phixu - cfg.gamma * phixpup)';
            b = b + phixu * Rs(i);
            if i/Ns >= progs,    % display progress inline
                dispx([8 8 8 8 sprintf('%2d%%', fix(i/Ns*100))], cfg.verb, 3);
                progs = progs + cfg.sampledisp;
            end;
        end;
        dispx([8 8 8 8 '100%.'], cfg.verb, 3);
        timestat.iter_computeAb(k) = cputime - tmark; tmark = cputime;
        % b) solve the least-squares problem
        dispx([8 ' Solving LSQR...'], cfg.verb, 3);
        lastwarn('');
        % theta = A \ b;    
        % replaced with version w/ normalization on 2009-09-10
        theta = (A./Ns) \ (b./Ns);
        if ~isempty(lastwarn), Anoninv = 1; end;    % set non-invertibility flag 
        dispx([8 ' done.'], cfg.verb, 3);        
        timestat.iter_qsolve(k) = cputime - tmark; tmark = cputime;
        
        % Perform policy improvement
        % a) process samples, computing the RHS of the least-squares eqn
        progs = cfg.sampledisp; dispx(sprintf('k=%d: h improv., samples processed: %2d%%', k, 0), cfg.verb, 3);
        hb = zeros(hNs, 1);
        for i = 1:hNs,
            hb(i) = approx.h(approx, theta, hXs(:, i));
            if i/hNs >= progs,    % display progress inline
                dispx([8 8 8 8 sprintf('%2d%%', fix(i/hNs*100))], cfg.verb, 3);
                progs = progs + cfg.sampledisp;
            end;
        end;
        dispx([8 8 8 8 '100%.'], cfg.verb, 3);
        timestat.iter_computeh(k) = cputime - tmark; tmark = cputime;
        % b) solve the least-squares problem
        if isempty(con),    % b1) unconstrained
            dispx([8 ' Solving LSQR...'], cfg.verb, 3);
            lastwarn('');
            htheta = HPHI \ hb;
            HPHIpoor = ~isempty(lastwarn);      % set poorly-conditioned flag 
            dispx([8 ' done.'], cfg.verb, 3);
        else                % b2) constrained
            dispx([8 ' Solving constrained LSQR...'], cfg.verb, 3);
            [htheta,resnorm,residual,exitflag,output] = lsqlin(HPHI, hb, Acon, bcon, ...
                [], [], [], [], htheta, optcon); % be smart and use the last htheta as start point
            HPHIpoor = 0;
            dispx([8 sprintf(' done in %d iter, exitflag=%d.', output.iterations, exitflag)], cfg.verb, 3);
        end;
        timestat.iter_hsolve(k) = cputime - tmark; tmark = cputime; % stats not counted
        
        % add Q-function parameter vector to history, compute delta
        thetah{k+1} = theta; 
        deltah(k+1) = norm(theta - thetah{k}, 2);
        % save policy parameter on history, compute delta
        hthetah{k+1} = htheta;
        hdeltah(k+1) = norm(htheta - hthetah{k}, 2);

        % check convergence condition 
        conv = hdeltah(k+1) <= cfg.eps;
        if cfg.checkq, conv = conv & (deltah(k+1) <= cfg.eps); end;
        
        % check for oscillations
        per = 2; osc = 0;   % don't test for period = 1, that's taken care of by the simple convergence test
        % try periods up to maxper, 
        % as long as oscillation wasn't detected (then it makes no sense to try longer periods)
        % and provided that the number of iterations allows for it
        while ~osc && (per <= cfg.maxper) && (per * (cfg.nper+1) + 1 < k+1)
            osc = 1;                % start by assuming oscillating; tests below might falsify that
            for i = 0:per-1,                % i loops through the offsets of the iters that have to be equal
                for j = 1:cfg.nper,                % make sure oscillation happens for at least nper periods
                    osc = osc & all(abs(hthetah{k+1-i} - hthetah{k+1-j*per-i}) <= cfg.epsper);
                    % also verify value fun parameters if required
                    if cfg.checkq, osc = osc & all(abs(thetah{k+1-i} - thetah{k+1-j*per-i}) <= cfg.epsper); end;
                    % two-norm check -- saved in case in need of using it
                    % osc = osc & (norm(hthetah{k+1-i} - hthetah{k+1-j*per-i}) <= cfg.epsper);
                    if ~osc, break; end;            % don't test farther periods
                end;
                if ~osc, break; end;        % don't test other offets
            end;
            per = per + 1;
        end;
        per = per - 1;                      % revert the last increment
        if osc, kosc = k; conv = -1; end;   % flag oscillation

        timestat.iter_convtest(k) = cputime - tmark; clear tmark;
        
        % update aggregate statistics
        timestat.iter_peval(k) = timestat.iter_computeAb(k) + timestat.iter_qsolve(k);
        timestat.iter_pimpr(k) = timestat.iter_computeh(k) + timestat.iter_hsolve(k);
        timestat.iter_run(k) = timestat.iter_peval(k) + timestat.iter_pimpr(k) + timestat.iter_convtest(k);
        timestat.run = timestat.run + timestat.iter_run(k);
        trun = timestat.run;                % keep trun for compatibility reasons
        
        % compute and store condition number for A
        if cfg.usesparse, condA(k) = condest(A); else condA(k) = cond(A); end;
        % store A and b if configured to do so
        if cfg.storeab, Ah{k} = A; bh{k} = b; end;
        
        % visual feedback of algorithm progress
        if ~mod(k, cfg.iterdisp),
            dispx(['*** k=' num2str(k) ' LSPI iter done, cond(A)=' num2str(condA(k))...
                ' delta_k+1=' num2str(deltah(k+1))], cfg.verb, 2);
        end;
        % data backup
        if ~mod(k, cfg.itersave),
            save(cfg.datafile);
            dispx(['Data at iter k=' num2str(k) ' saved to [' cfg.datafile '].'], cfg.verb, 1);
        end;
        
        k = k + 1;
    end;        % WHILE not converged and allowed more iterations

    if conv == 1,   dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    elseif osc,     dispx(['Solution oscillates with period ' num2str(per) '. Algorithm stopped'], cfg.verb, 0);
    elseif Anoninv, dispx('The A matrix is singular or badly conditioned. Algorithm stopped.', cfg.verb, 0);
    elseif HPHIpoor,dispx('The HPHI matrix is badly conditioned. Algorithm stopped.', cfg.verb, 0);
    else            dispx(['maxiter=' num2str(cfg.maxiter) ' exhausted. Algorithm stopped'], cfg.verb, 0);
    end;
    
end;


% Options below might create figures, init figure handles array
figh = [];

% ==== REPLAY POLICY ====
if cfg.replay,

    % initial state
    if ~isempty(cfg.x0), 
        if ischar(cfg.x0) && ~isempty(cfg.problem),
            % use the model function to get a string-named initial state
            x0 = feval(cfg.problem, 'x0', cfg.x0);
        else
            x0 = cfg.x0(:);         % specified initial state
        end;
    else    % try getting default initial state
        try         x0 = feval(cfg.problem, 'x0');
        catch       x0 = zeros(model.p, 1);    % zeros
        end;
    end;         
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(model.p, length(t)); x(:, 1) = x0;
    u = zeros(model.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    for k = 1:Ns,
        % compute action
        u(:, k) = happrox.h(happrox, htheta, x(:, k));
        % apply to system
        [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
        if terminal, Ns = k; u(:, k+1) = NaN; break; end;      % entered terminal state
    end;
 
    % plot history & optionally save figures
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);
    hist.R = discreturn(cfg, hist.r, Ns, terminal);
    if ~cfg.silent,
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        else
            figh(end+1) = plothistory(hist);
        end;
%         % also put the return in the figure title
%         title(['R(x_0)=' num2str(hist.R(1))]);
        setfigprop(cfg);
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay

% ==== STATISTICS ====
if (cfg.evol || cfg.sol) && ~cfg.silent,
    % grayscale styles
    gs.cm = gray(96); gs.cm = gs.cm(24:end);
    % color styles
    cs = gs;
    cs.cm = jet(64);
    % set style
    if cfg.grayscale, sty = gs; else sty = cs; end;    
    
    % readable labels
    labels.iter = 'Iteration'; 
    labels.h = 'h(x)'; labels.V = 'V(x)'; 
    labels.delta = '$\Vert \theta_\ell - \theta_{\ell-1}\Vert_2$'; 
    labels.hdelta = '$\Vert \vartheta_\ell - \vartheta_{\ell-1}\Vert_2$'; 
    labels.condA = 'cond(A)';
    latexprop = {'Interpreter', 'Latex', 'FontSize', 11};
end;

if cfg.sol && ~cfg.silent,
    if exist('Anoninv', 'var') && Anoninv,
        dispx('A was non-invertible, solution is meaningless. Skipping plot.', cfg.verb, 0);
    else
        figh(end+1) = figurex([1100 450]); % colormap(sty.cm);
        K = find(~isnan(deltah), 1, 'last') - 1;
        % value function
        subplot(3, 2, [1 3]); cla;
        approx.plotv(approx, theta); title(labels.V);
        subplot(3, 2, [2 4 6]); cla;
        % policy
        happrox.ploth(happrox, htheta, 'npoints=100'); title(labels.h);
        % deltas for Q and h
        subplot(3, 2, 5); cla;
        semilogy(1:K, deltah(2:K+1), 'k'); grid on; hold on;
        semilogy(1:K, hdeltah(2:K+1), 'k'); xlabel(labels.iter);
        lh = legend(labels.delta, labels.hdelta); set(lh, latexprop{:});
        setfigprop(cfg);
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;
    
if cfg.evol && ~cfg.silent,
    if exist('cleanedup', 'var') && cleanedup >= 2,
        dispx('Cannot replay evolution, hard cleanup was performed on the datafile.', cfg.verb, 0);
    elseif exist('Anoninv', 'var') && Anoninv,
        dispx('A was non-invertible. Skipping evolution plot.', cfg.verb, 0);
    else
        figh(end+1) = figure; colormap(sty.cm);
        K = find(~isnan(deltah), 1, 'last') - 1;
        for k = 1:K+1,
            figcfg.figname = sprintf('Iter#%d, %s', k, cfg.datafile);
            setfigprop(figcfg);
            subplot(221); cla;
            approx.plotv(approx, thetah{k}, 'npoints=30'); title(labels.V);
            subplot(222); cla;
            happrox.ploth(happrox, hthetah{k}); title(labels.h);
            subplot(223); cla;
            semilogy(1:k, deltah(2:k+1), 'k'); grid on; hold on;
            semilogy(1:k, hdeltah(2:k+1), 'k'); xlabel(labels.iter);
            lh = legend(labels.delta, labels.hdelta); set(lh, latexprop{:});
            if k <= K && exist('condA', 'var'),   % A not defined after the final iteration
                subplot(224); cla;
                semilogy(1:k, condA(1:k), 'k'); xlabel(labels.iter); ylabel(labels.condA);
            end;
            if k <= K, pause; end;
        end;
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;

if ~isempty(cfg.exth) && ~cfg.silent,
    EXTH.select = 'h';     % policy param ('h') or Q-function param ('Q')
    EXTH.nvar = 10;        % how many last theta's to consider when choosing the one that varies most
    EXTH.n = 5;            % how many params to plot
    % if exth just a boolean,leave defaults
    if ~isstruct(cfg.exth) && ~ischar(cfg.exth), cfg.exth = EXTH; end;
    exth = parseconfig(cfg.exth, EXTH);
    K = find(~isnan(deltah), 1, 'last') - 1;
    % make thetah matrix out of last nvar theta-s; compute variance across iterations
    % (ignore the first, all-zeros theta vector)
    if exth.select == 'h',  thetavar = cell2mat({ hthetah{1 + max(1, K-exth.nvar+1):K} });
    else                    thetavar = cell2mat({ thetah{1 + max(1, K-exth.nvar+1):K} });
    end;
    thetavar = var(thetavar, 0, 2);
    % find indices of n largest elements in thetavar
    itheta = findnlargest(thetavar, exth.n);
    % collect the values of thetah(itheta) for every k
    vtheta = zeros(exth.n, K);
    for k = 1:K, 
        if exth.select == 'h',  vtheta(:, k) = hthetah{k+1}(itheta); 
        else                    vtheta(:, k) = thetah{k+1}(itheta); 
        end;
    end;
    % plot them 
    % note these plotting styles are only a quick patchwork
    sty.lines = {'k-', 'k--', 'k:'};
    figh(end+1) = figure;
    for i = 1:size(vtheta, 1),
        plot(vtheta(i, :), sty.lines{i}, 'Marker', 'o', 'MarkerSize', 3, 'MarkerFaceColor', 'k'); hold on;
    end;
    xlabel('Iteration');
    if exth.select == 'h',
        ylabel('\vartheta');
    else
        ylabel('\theta');
    end;
    setfigprop(cfg); % set(gcf, 'NumberTitle', 'off', 'Name', cfg.datafile);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% ==== BACKUP DATA ====
if cfg.run || cfg.resume, 
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir; 
        if any(datadir(end) == '\/'), datadir = datadir(1:end-1); end;
    end;
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        if exist([CFG.datafile '.mat'], 'file'), delete([CFG.datafile '.mat']); end;
    elseif ~strcmp(datadir, pwd)
        delete([cfg.datafile '.mat']);   % need to save in a different directory, so delete anyway
    end;    % otherwise just overwrite
    cfg.datafile = [datadir '/' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% set output
if cfg.run || cfg.resume,       % output optimal solution
    varargout = {theta, htheta};
elseif cfg.replay,              % output history and possibly figure handles
    varargout = {hist, figh};
elseif cfg.evol || cfg.sol || ~isempty(cfg.exth),               % output fig handles (replay takes precedence)
    varargout = {figh};
end;

end
% lspi() RETURNING varargout =================================================================
