function varargout = holop(x0, model, cfg)
% HOLOP Planning with OLOP over K-horizon sequences
%   U0 = HOLOP(X0, MODEL, CFG)
%   [U0, PLANINFO] = HOLOP(X0, MODEL, CFG)
%   [U0, PLANINFO, PLANDATA] = HOLOP(X0, MODEL, CFG)
%   [USEQ, XSEQ, PLANINFO, PLANDATA] = HOLOP(X0, MODEL, CFG)
% Implementation of algorithm from: (also looking at the code)
%   Ari Weinstein and Michael Littman, "Bandit-Based Planning and Learning in
%   Continuous-Action MDPs", Proceedings 22nd ICAPS, 2012.
% Differences: removed bug in nu1 init

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % simulation budget
CFG.K = [];                     % planning horizon / rollout length
CFG.detact = 1;                 % whether the algo should sample actions, and break action ties, deterministically
                                % (note dimension selection is still random regardless of this)
CFG.k = -1;                     % discrete time step
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% parameter and error checks
if model.q > 1, error('HOLOP currently supports only scalar actions'); end;

% start counting time
t = cputime;

% quick-access to config parameters, and other helper vars
maxu = model.maxu; detact = cfg.detact; K = cfg.K;
discounting = cfg.gamma .^ (0:K-1);
% CDF of probability distribution with which dimensions to split are chosen
dimcdf = cumsum(discounting ./ sum(discounting)); dimcdf = [0 dimcdf(1:end-1)];

% hardcoded parameters, chosen as in the original implementation (and as suggested in OLOP paper)
rho = 2^(-2/K);
nu1 = (sqrt(K)/2)^2;

% 0) init counts and data structures
maxrollouts = ceil(cfg.n / K);% number of rollouts
treesize = 1+2*maxrollouts;   % at each rollout we create two new nodes
% tree
TP = NaN(3, treesize);        % tree pointers: first is parent, next two are the children
TU = NaN(K, treesize);        % start points of domains, for each depth and node
TE = NaN(K, treesize);        % end points
TN = NaN(1, treesize);        % sample counts, one for each node
TR = NaN(1, treesize);        % cumulative sum of truncated returns
TB = NaN(1, treesize);        % B-value
TD = NaN(1, treesize);        % depth
TLEAF = NaN(1, treesize);     % leaf flags

% start by initializing a root node with the entire domain and no samples
TP(:, 1) = 0;
TU(:, 1) = 0; TE(:, 1) = 1;
TN(1) = 0; TR(1) = 0; TB(1) = Inf;
TD(1) = 0; TLEAF(1) = true;
N = 1;              % number of nodes so far

% rollouts
Xh = NaN(model.p, K+1, maxrollouts);
UNORMh = NaN(K, maxrollouts);         % normalized U; recall q = 1
Uh = NaN(K, maxrollouts); 
Rh = NaN(K+1, maxrollouts);
RETh = NaN(1, maxrollouts);           % returns
nrollouts = 0;

%#### Main loop
stopflag = false;
while ~stopflag,

    % 1) sample HOO 
    % navigate the existing tree along optimistic path (remembering path)
    node = 1; optimpath = node;
    while ~TLEAF(node),
        children = TP(2:3, node);
        % pick children with maximal B-value, breaking ties determ or random, as configured
        node = children(max(TB(children)) == TB(children));
        if ~isscalar(node), if detact, node = node(1); else node = node(unidrnd(2)); end; end;
        optimpath(end+1) = node;
    end;
    optimleaf = node;

    % choose a sample sequence within the node
    nrollouts = nrollouts + 1;
    if detact,  % sample middle
        UNORMh(:, nrollouts) = TU(:, optimleaf) + 0.5 .* (TE(:, optimleaf) - TU(:, optimleaf));
    else        % sample randomly, making sure we don't hit the right edges of the domain
        r = rand(K, 1); if any((TE(:, optimleaf) < 1) & (r == 1)), r = rand(K, 1); end;
        UNORMh(:, nrollouts) = TU(:, optimleaf) + r .* (TE(:, optimleaf) - TU(:, optimleaf));
    end;
    
    % 2) do a rollout with the chosen sequence
    Xh(:, 1, nrollouts) = x0;
    Uh(:, nrollouts) = unnormu(UNORMh(:, nrollouts), maxu);     % unnormalized actions
    for k = 1:K,            
        [Xh(:, k+1, nrollouts), Rh(k+1, nrollouts), term] = ...
            feval(model.fun, model, Xh(:, k, nrollouts), Uh(k, nrollouts));
        if term, error('HOLOP does not yet support terminal states'); end;
    end;
    % compute the truncated return -- this is the value of the action sample
    RETh(nrollouts) = discounting * Rh(2:end, nrollouts);

    % 3) add rollout information to the tree (increase sample counts, accumulate rewards)
    % note new children below do not inherit this info, but recompute it by looking for the actual
    % samples
    TN(optimpath) = TN(optimpath) + 1;
    TR(optimpath) = TR(optimpath) + RETh(nrollouts);

    % 4) expand optimistic leaf, creating its two new children
    dsplit = find(rand >= dimcdf, 1, 'last');   % split
    children = [N+1 N+2];
    for i = 1:2,
        child = children(i);
        % update parent
        TP(i+1, optimleaf) = child;     % create pointer towards this child node
        TLEAF(optimleaf) = false;
        % create child
        TP(:, child) = [optimleaf 0 0]; % create pointer towards parent; no children exist yet
        TU(:, child) = TU(:, optimleaf);
        TE(:, child) = TE(:, optimleaf);
        if i == 1,  % left
            TE(dsplit, child) = TU(dsplit, child) + (TE(dsplit, child) - TU(dsplit, child)) / 2;
        else        % right
            TU(dsplit, child) = TU(dsplit, child) + (TE(dsplit, child) - TU(dsplit, child)) / 2;
        end;
        % child may already contain some samples: find them (as in original implementation)
        isamples = findnodesamples(child, TU, TE, UNORMh, nrollouts);
        TN(child) = nnz(isamples); TR(child) = sum(RETh(isamples)); 
        TD(child) = TD(optimleaf) + 1; 
        TLEAF(child) = true;
    end;
    N = N+2;        % increment nodes counter

    % 5) compute B-values
    % own B-values
    TB(find(TN(1:N) == 0)) = Inf;   % work only with initialized part rather than the whole arrays
    sampled = find(TN(1:N) > 0);
    TB(sampled) = TR(sampled) ./ TN(sampled) + sqrt(2 * log(nrollouts) ./ TN(sampled)) + nu1 * rho.^TD(sampled);
    % propagate upwards into the tree, starting from max depth - 1 (last depth are surely leaves)
    for d = max(TD(1:N))-1:-1:0,
        % work with inner nodes at this depth: leaves keep their own b-values
        innernodes = find(TD(1:N) == d & ~TLEAF(1:N));
        if isempty(innernodes),          % no nodes
            continue;
        elseif isscalar(innernodes),    % the easy case
            TB(innernodes) = min(TB(innernodes), max(TB(TP(2:3, innernodes))));
        else
            % TP(2:3, nodes) is a 2 x #innernodes matrix of child indices
            % TB(this) are their b-values, in the same structure
            % maximum along columns => row vector, then minimum with parent which is also row vector
            TB(innernodes) = min(TB(innernodes), max(TB(TP(2:3, innernodes)), [], 1));
        end;
    end;
    
    % stop when the budget has been exhausted: 
    % note we may use more than the allowed budget to perform the last rollout
    stopflag = (nrollouts == maxrollouts);
    
end;    % end main tree construction loop

% we should have exactly as many nodes as we predicted
if N ~= treesize, error('Unexpected tree size, %d instead of %d', N, treesize); end;

% 4) Find best action sequence (as in original implementation)
% navigate tree along greedy path
Rhat = -Inf + zeros(1, N);      % default values are -Inf, this is the value remaining for unsampled nodes
sampled = find(TN(1:N) > 0);
Rhat(sampled) = TR(sampled) ./ TN(sampled);
node = 1;
while ~TLEAF(node),
    children = TP(2:3, node);
    node = children(max(Rhat(children)) == Rhat(children));
    % tie breaking
    if ~isscalar(node), if detact, node = node(1); else node = node(unidrnd(2)); end; end;
end;
% choose sample with best return within this node
% note this choice is not random, but first-found (according to original implementation)
isamples = find(findnodesamples(node, TU, TE, UNORMh, nrollouts));
[Rstar, iistar] = max(RETh(isamples)); istar = isamples(iistar);
% we already have the unnormalized action samples -- make sure they have the right shape
useq = reshape(Uh(:, istar), model.q, K);    
xseq = Xh(:, :, istar);

% create stats if requested on output
if nargout >= 2,
    planinfo.n = nrollouts * K;
    planinfo.nrollouts = nrollouts;
    planinfo.hoodepth = max(TD(1:N));
    planinfo.hootreesize = N;
    planinfo.trun = cputime - t;
    planinfo.message = sprintf('Planning done, n=%d transitions, nrollouts=%d (HOO tree depth=%d, size=%d), trun=%.2f', ...
        planinfo.n, planinfo.nrollouts, planinfo.hoodepth, planinfo.hootreesize, planinfo.trun); 
end;

% create plandata if requested on output
if nargout >= 3,
    plandata = varstostruct(...
        'TP', 'TU', 'TE', 'TN', 'TR', 'TB', 'TD', 'TLEAF',...
        'Xh', 'UNORMh', 'Uh', 'Rh', 'RETh', ...
        'useq', 'xseq');
end;

% construct the output based on calling mode
switch nargout,
    case 1, varargout = {useq(:, 1)};
    case 2, varargout = {useq(:, 1), planinfo};
    case 3, varargout = {useq(:, 1), planinfo, plandata};
    case 4, varargout = {useq, xseq, planinfo, plandata};
end;

end % main function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
% normalize / unnormalize actions

% function u = normu(u, maxu), u = 0.5 + 0.5 * u / maxu; end

function u = unnormu(u, maxu)
u = (u - 0.5) * 2 * maxu; 
end

% find indices of all rollout samples inside a node's action range
% returns a logical row vector with size nrollouts
function isamples = findnodesamples(node, TU, TE, UNORMh, nrollouts)
% quickreturn if no rollouts
if nrollouts == 0, isamples = false(nrollouts, 1); return; end;
% vectorized inclusion test
uext = repmat(TU(:, node), 1, nrollouts);
eext = repmat(TE(:, node), 1, nrollouts);
usamples = UNORMh(:, 1:nrollouts);
% note intervals are open to the right to avoid overlaps, except the last interval which also
% includes the end-point = 1
isamples = all(...
    (usamples >= uext) & ...
        ( ((eext < 1) & (usamples < eext)) | ((eext == 1) & (usamples <= eext)) ), ...
    1);
end
