function out = learn_lipschitz(mode, cfg, approx, x, v)
% Implements the Lipschitz based approximator for OP
% Currently can be used in combination with OPMDP
% Required parameters on the config:
%   gamma               - discount factor
%   dimx (set automatically in the planner)
%   lipschitz           - Lipschitz constant
%   mem_maxsize          - maximum memory size

if mode(1) == 'i',      % 'init' mode
    % initialization consists of creating an empty memory
    approx = struct;
    approx.x = NaN(cfg.dimx, 0, 'double');   % state
    approx.v = NaN(1, 0, 'double');            % values
    approx.memsize = 0;                        % memory initially empty
    
    out = approx;
    
elseif mode(1) == 'e',  % 'estimateb' mode
    % quick-return for empty approximator 
    if ~approx.memsize, out = 1/(1-cfg.gamma); return; end;
    
    % compute L1 distance to all points in memory, dimension-wise
    dist = zeros(1, approx.memsize);
    for ii = 1:cfg.dimx, dist = dist + abs(x(ii) - approx.x(ii, 1:approx.memsize)); end;
    % find value estimates given by each of these points
    v = approx.v(1:approx.memsize) + cfg.lipschitz .* dist;
    % choose the minimum (and a default 1/1-gamma if the minimum is larger)
    v = min(min(v), 1/(1-cfg.gamma));
    
    out = v;

elseif mode(1) == 'u',  % 'update' mode
    
    
    % 1) Aggregate duplicates
    % start with existing memory and values on the inner tree nodes
    % (leaves are not provided by the main function, since they don't add information)
    [Mx, Mv] = learn_removeduplicates([approx.x, x], [approx.v, v], cfg);
    memsize_initial = approx.memsize+length(v);
    memsize = size(Mx, 2);
    memsize_duplicates = memsize_initial - memsize;
    
    % 2) Remove redundant points (under the assumption of L1 norm)
    Ms = zeros(1, memsize);
    for i = 1:memsize,
        vi = Mv(i);         % save value of point i
        Mv(i) = Inf;        % equivalent to disregarding this point when estimating value
        % compute the estimated value of the point WITHOUT it in the database
        dist = zeros(1, memsize);
        for ii = 1:cfg.dimx, dist = dist + abs(Mx(ii, i) - Mx(ii, :)); end;
        v = Mv + cfg.lipschitz .* dist;
        v = min(min(v), 1/(1-cfg.gamma));
        % score of each point is the difference between the estimated bound /without/ the point in
        % the set, and its own bound;
        % this score is negative when the sample is redundant, i.e. its bound is larger than
        % resulting from neighbors (or than absolute bound 1/1-gamma)
        Ms(i) = v - vi;
        Mv(i) = vi;         % restore value of point i
    end;
    % points that do not contribute to the aggregate bound are redundant
    % so are points larger than the absolute upper bound 1/1-gamma
    keep = Ms > 0;    
    % shrink memory
    Mx = Mx(:, keep);
    Mv = Mv(keep);
    memsize_redundant = memsize - size(Mx, 2);
    memsize = size(Mx, 2);

    % 3) Remove other points sorted by their score to shrink to allowed memory size
    if cfg.mem_maxsize > 0 && memsize > cfg.mem_maxsize,
        % the aggregate score is the maximum of scores for each action
        Ms = Ms(keep);
        [Ms, irearrange] = sort(Ms, 'descend');
        Mx = Mx(:, irearrange(1:cfg.mem_maxsize));
        Mv = Mv(irearrange(1:cfg.mem_maxsize));
        memsize_pruned = memsize - cfg.mem_maxsize;
        memsize = size(Mx, 2);
    else
        memsize_pruned = 0;
    end;

    approx.x = Mx;
    approx.v = Mv;
    approx.memsize = memsize;
    
    out = approx;
    
    if cfg.verb >= 3,
        disp(sprintf('Updated memory, memsizes:init=%d/dup=%d/redund=%d/pruned=%d/fin=%d', ...
            memsize_initial, memsize_duplicates, memsize_redundant, memsize_pruned, memsize));
    end;

    
else
    error('Unknown mode [%s]', mode);
end;