function rtrobot_init(Port,ref,Kp)
% initialize the Ed-Ro robot to a desired pose (0 by default)
%    RTROBOT_INIT(Port,ref,Kp)
%
%       Port ... port, e.g., 'COM1' (default Port = 'COM1')
%       ref  ... reference to initialize to (default ref = 0)
%       Kp   ... proportional gain (default Kp = 3)

% (c) Robert Babuska, 2005

if nargin < 1, Port = 'COM8'; end;              % default port
if nargin < 2, ref = 0; end;                    % default reference
if nargin < 3, Kp = 5; end;                     % default control gain

h = 0.05;                                       % sampling period
eps = 0.05;                                     % tolerance around ref
umax = 0.5;                                     % maximum allowed control 

% initialize, open the port, define timer
hwinit;                                         % get HW parameters
n = length(joint);                              % number of joints to be controlled
joint = joint(:)+(0:6:6*(n-1))';                % convert into sequential index
joint_select = zeros(6,n);                      % prepare zeros
joint_select(joint) = 2;                        % assing 1 at the right places

com = rs232('GetParams','default');             % get default RS232 parameters
com.Port = Port;                                % define port number
com.BaudRate = 115200;                          % define Baud rate
com.ReadTimeout = 1;                            % define short timeout
com.WriteTimeout = 1;                           % define short timeout
[params,result]=rs232('open',com);              % open the port

xx = zeros(2,7);                                % aux var for BE conversion

% control loop
tic;
for i = 1 : 100,                                % run for at most 100*h seconds
    rs232('write',com,uint8(128));              % control byte for read D/A
    xx(:) = rs232('read',com,14);               % read 14 bytes from 7 sensors
    x = double(xx'*[256 1]');                   % convert to big endian
    x = adingain(:).*(x + adinoffs(:));         % scale to physical range
    x = joint_select'*x(1:6);                   % select the right variable
    if all(abs(ref-x) < eps); break; end;       % if sufficiently close, stop
    u = min(umax,max(-umax,Kp.*(ref-x)));       % saturate control action
    u = joint_select*u;                         % complete control vector
    u = daoutgain(1:6).*u' + daoutoffs(1:6);    % scale to digital range
    rs232('write',com,uint8(0));                % control byte for send all actuators
    rs232('write',com,uint8(u)');               % write 6 bytes to 6 actuators
    while toc < h, end;                         % synchronize with real time
    tic;                                        % reset Matlab's tic-toc timer    
end;

% reset actuators, close the port
rs232('write',com,uint8(56));                   % laser off
rs232('write',com,uint8(0));                    % control byte for send all actuators
rs232('write',com,uint8(127*ones(6,1)));        % reset all 6 actuators to 0 (127 bin)
rs232('close',com);                             % close serial port
pause(.2);                                      % wait a while