function [xplus, rplus, terminal, pplus] = hillcar_mdp(m, x, u)
% Implements the discrete-time dynamics of the hillcar.
%  [XPLUS, RPLUS, TERMINAL] = HILLCAR_MDP(M, X, U)
%
% This function conforms to the specifications established by SAMPLE_MDP.

% WARNING: this function does NOT check if X is terminal, so algorithms should take care of that and
% not continue asking for transitions once a terminal state has been reached.

% REMARK: after checking if x has gone over the bounds thus leading to a terminal state, x+ is
% actually bounded to the closed domain, which means it is NOT technically terminal. This is for the
% sake of the approximators, which may sometimes be called to compute approximate values for
% terminal states, and may lead to unexpected results when the state is not part of the closed
% domain. This behavior is not a problem if terminal states are properly handled, since the terminal
% reward is unchanged, there is no next Q-value, and the algorithm should stop regardless of the
% value of the next state.
% Note that even when X0 contains values with x at the bound, and simulations are started from such
% states, there is no issue, since those states are NOT terminal.
% Starting from 2011-07-06, bounding of terminal states can be turned off from config field

% BUGFIX: 2011-07-06: Fixed bug where +1 reward condition was 
% x(1) > m.maxx(1) && x(2) <= m.maxx(2); should be abs(x(2) <= m.maxx(2)). Since it is physically
% impossible to exit to the right with velocity exceeding negative bound, the bug had no effect

% numerical integration up to next state
if m.disc.method == 2,          % discrete ODE
    odey = feval(m.disc.odesolver, @trans, m.disc.odet, x, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts
        
elseif m.disc.method == 1,      % ODE
	[odet odey] = feval(m.disc.odesolver, @trans, m.disc.odet, x, m.disc.odeopt, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts

elseif m.disc.method == 3,      % Euler, x(k+1) = x(k) + Ts f(x(k), u(k))
    dt = m.disc.Ts / m.disc.odesteps; xplus = x;
    for i = 2:m.disc.odesteps,
        xplus = xplus + dt * trans(0, xplus, u);
    end;
    
end;

% terminal state handling
terminal = any(abs(xplus) > m.maxx);    % terminal state
success = terminal && xplus(1) > m.maxx(1) && abs(xplus(2)) <= m.maxx(2);

% reward computation
if ~isfield(m, 'rewtype') || m.rewtype(1) == 't',   % 'term'ination reward only, -1 for failure
    if ~terminal,       rplus = 0;      % regular state
    elseif success,     rplus = 1;      % terminated with success
    else                rplus = -1;     % failure: terminal but no sucess
    end;
elseif m.rewtype(1) == 'p',                         % 'prog'ress based, 0 for failure
    if ~terminal,       rplus = m.progrew * (1 - (m.maxx(1) - xplus(1)) / 2*m.maxx(1));
    elseif success,     rplus = 1;
    else                rplus = 0;
    end;
else
    error('Unknown reward type [%s]', m.rewtype);
end;

if terminal && (~isfield(m, 'boundxterm') || m.boundxterm),
    % bound the terminal state (for the sake of the approximators)
    xplus = max(-m.maxx, min(m.maxx, xplus));
end;

if nargout == 4, pplus = 1; end;    % output probability

% END FUNCTION hillcar_mdp() RETURNING xplus, rplus, terminal -----

function xdot = trans(t, x, u)
%  Implements the transition function of the car on the hill problem
%
%  Parameters:
%   T           - the current time. Ignored, added for compatibility with
%               the ode function style.
%   X           - the current state ([position; velocity]
%   U           - the command
%
%  Returns:
%   XDOT        - the derivative of the state
%

p = x(1);
if p < 0,
    h1 = 2*p+1; h2 = 2;
else
    h1 = (1+5*p*p)^-1.5; h2=-15*p/(1+5*p*p)^2.5;
end;
% below, mass assumed 1, gravitational constant 9.81
xdot = [x(2); (u - 9.81*h1 - x(2)^2*h1*h2) / (1+h1*h1)];
