function [xplus, rplus, terminal] = invpend_mdp(m, x, u)
% Implements the discrete-time dynamics of an inverted pendulum
%  [XPLUS, RPLUS, TERMINAL] = INV_MDP(M, X, U)
%
% This function conforms to the specifications established by SAMPLE_MDP.

u = sat(u, -m.maxu, m.maxu);

if m.odemethod(1) == 'v',          % Variable-step ODE
	[odet odey] = m.odesolver(m.trans, m.odet, x, m.odeopt, m, u);
    xplus = odey(end, :)';

elseif m.odemethod(1) == 'f',      % Fixed-step ODE
    odey = m.odesolver(m.trans, m.odet, x, m, u);
    xplus = odey(end, :)';

elseif m.odemethod(1) == 'e',      % Euler, x(k+1) = x(k) + Ts * f(x(k), u(k)) (possibly multistep)
    dt = m.Ts / m.odesteps; xplus = x;
    for i = 1:m.odesteps,
        xplus = xplus + dt * m.trans(0, xplus, m, u);
    end;
end;

% wrap angle in [-pi, pi), saturate velocity
xplus = [mod(xplus(1) + pi, 2*pi) - pi; sat(xplus(2), -m.maxx(2), m.maxx(2))];

% Reward
if m.rewtype(1) == 'l',         % 'lqr'
    rplus = - x' * m.Q * x - m.R * u * u;
    terminal = 0;   % task is continuing
elseif m.rewtype(1) == 'g',     % gauss
    rplus = -1 + exp(-[x; u]' * m.W * [x; u]);
    terminal = 0;
elseif m.rewtype(1) == 'b',    % box
    rplus = all(abs(x) <= m.boxrew);
    terminal = 0;
elseif m.rewtype(1) == 'f',     % 'fail'
    terminal = m.zeroreward * abs(xplus(1)) > pi/2;
    xplus(1) = max(min(xplus(1), pi/2), -pi/2);
    rplus = -terminal;
end;

end
% END FUNCTION int_mpd() RETURNING xplus, rplus ==================================================
