function [trajhist, figh] = ipsetup_control(cfg)
% Control of the physical IP setup.

% run config
CFG.stype = 'ip';   % flag which setup we are using
CFG.p = 2;              % order of the system
CFG.Tend = 5;           % time to run
CFG.Ts = 0.005;         % sampling period
CFG.maxu = 3;           % maximum allowed control 
CFG.maxomega = 15*pi;    % maximum angular velocity
CFG.htype = 'fuzzy'; 
CFG.datafile = 'D:\Work\m\approxrl\systems\invpend_setup\ipconsstudy\is_lqr_N41M15';
CFG.savedatadir = [pwd '\'];
CFG.savedatafile = [];
CFG.x0 = 'down';  
CFG.randomX0 = 0;       % old-style random initial state (takes precedence over x0)
CFG.trandc = 1.5;
CFG.keeprandc = 20;
CFG.interph = 0;        % interpolate fuzzy policy
CFG.plotmodel = 0;
CFG.silent = 0;

if nargin < 1, cfg = struct; end;
cfg = parseconfig(cfg, CFG);
if cfg.randomX0, cfg.x0 = 'rand'; end;  % old-style random x0 takes precedence
structtovars(cfg);  % for quick/realtime access

K = Tend/Ts;        % # of steps to run

% ===================================
% Initialize control policy
switch htype,
    case 'fuzzy',
        FZH = fuzzyqi_prepareh(datafile, interph); 
    case 'lspi',    % note LSPIH not yet supported!
        loadwithprefix(datafile, 'lspi_', {'approx', 'theta'});
        LSAPP = revise_approx(lspi_approx);
        LSTHETA = lspi_theta;
        clear lspi_approx lspi_theta
    case 'step',
    case 'heur',
    case 'ramp',
    case 'pid',
        ref = pi/2;
        KP = 21.17;
        KD = 0.651;
        KI = 0.5;
        ierr = 0;
end;

% get handles to low-level control functions
cfuns = ipsetup_problem('controlfuns');


% ===================================
% Initialize DCMOTOR - USB init
% sensors structure: [status, time, angle, velocity, current, voltage, dangle/dt, ???]
mops('CloseAll');                % close port
H = mops('Open', 'mops1');       % open port
H.WatchdogTimeout = 10;          % watchdog timeout
mops('SetParams', H);            % set the parameter
mops('Write', H, 1, 1, 0, 0);    % dummy write to sync interface board
mops('Write', H, 1, 1, 0, 0);    % reset position
mops('Write', H, 0, 1, 0, 0);    % end reset
for i = 1 : 100,                 % read a couple of times
    mops('Read',H);              % dummy read sensor data
end;

% History variables
t = 0:Ts:K*Ts;
DATA = zeros(8, K+1);       % data history
X = zeros(p, K+1);          % initialize state history
U = zeros(1, K);            % initialize input history

if x0(1) == 'd',           % pointing downwards
    [H, Xinit] = cfuns.stabilize(H, maxu, Ts);
elseif x0(1) == 'r',      % apply random actions
    [H, Xinit] = cfuns.randomcontrol(H, maxu + 2, Ts, cfg.trandc, cfg.keeprandc);
elseif x0(1) == 'u',      % pointing upwards
    [H, Xinit] = cfuns.swingup(H, maxu, maxomega, Ts);
elseif x0(1) == 'U',      % point upwards then apply some random actions
    [H, Xinit] = cfuns.swingup(H, maxu, maxomega, Ts);
    [H, Xinit] = cfuns.randomcontrol(H, maxu, Ts, 0.2, 10);
end;
    

% ============================
% MAIN CONTROL LOOP
tic; 	% reset Matlab's tic-toc timer
for k = 1:K,                            
    DATA(:, k) = mops('Read',H);        % read sensor data
    % current angle
    X(1, k) = DATA(3, k);
    % velocity = dangle/dt
    if k > 2,   X(2, k) = (DATA(3,k) - DATA(3,k-1)) / Ts; %DATA(2,k);
    else        X(2, k) = (DATA(3,k) - Xinit(2)) / Ts; %DATA(2,k);    
    end;
        
    % compute control action
    if htype(1) == 'f',            % fuzzy
        % transform angle so that [0,0] = upward eqn, and wrap inside [-pi,pi)
        U(k) = fuzzyqi_h( normalizex(X(:, k), maxomega), (k-1)*Ts, FZH ); 
    elseif htype(1) == 'l',        % LSPI
        U(k) = LSAPP.h( LSAPP, LSTHETA, normalizex(X(:, k), maxomega) );
    elseif htype(1) == 's',        % step
        if k*Ts > 0.5, U(k) = 3; 
        else U(k) = 0; 
        end;
    elseif htype(1) == 'r',        % ramp
        U(k) = k*Ts;
    elseif htype(1) == 'p',        % PID
        err = ref - X(1,k);
        if k > 2,   errm = ref - X(1,k-1);
        else        errm = ref;
        end;
        ierr = ierr + err;
        derr = (err - errm) / Ts;
        U(k) = KP * err + KI * ierr + KD * derr;                          
    elseif htype(1) == 'h',
        x = normalizex(X(:, k), maxomega);
        if abs(x(1)) > pi/2,
            U(k) = 3 * (sign(x(2)) >= 0);
        else
            U(k) = -10 * x(1) -1 * x(2);
        end;
    end;
	
    U(k) = sat(U(k), -maxu, maxu);      % saturate control action
    mops('Write',H,0,1,U(k),0);         % send control input to process
    
    while toc < Ts, end;                % synchronize with real time
    tic;                                % reset Matlab's tic-toc timer
end;

% measure last state
DATA(:, k+1) = mops('Read',H);        % read sensor data
X(1, k+1) = DATA(3, k+1);
X(2, k+1) = (DATA(3,k+1) - DATA(3,k)) / Ts; %DATA(2,k+1);

% stabilize so that setup remains pointing down
H = cfuns.stabilize(H, maxu, Ts);
% stop motor -- reset actuators, close the port
mops('Write', H, 0, 1, 0, 0.0);
mops('CloseAll');

% normalize process state: shift angle by pi, wrap into -pi,pi
XN = normalizex(X, maxomega);

% create a model object
m = ipsetup_problem('model', [], ...
    struct('type', stype, 'maxu', maxu), ...
    struct('Ts', Ts)); 
%     struct('Ts', Ts, 'odemethod', 'varode', 'odesolver', @ode45));

% post-compute rewards
R = nan(1, K+1);          % reward history
for k = 1:K, 
    R(k+1) = ipsetup_reward(m, XN(:, k), U(k), XN(:, k+1)); 
end;

% simulate model
XM = XN .* 0; XM(:, 1) = XN(:, 1); RM = nan(size(R));
for k = 1:K,                             
    [XM(:, k+1), RM(k+1)] = m.fun(m, XM(:, k), U(:, k));    % task is continuing, don't care about term
end;

% plot stuff (if not silenced)
if ~cfg.silent,
    figh = figure; clf;
    subplot(411); leg = {};
    plot(t, XN(1, :), 'k'); leg{end+1} = '\alpha'; 
    if cfg.plotmodel, 
        hold on; plot(t, XM(1, :), 'b'); leg{end+1} = '\alpha_m';
    end;
    legend(leg{:});
    subplot(412); leg = {};
    plot(t, XN(2, :), 'k'); leg{end+1} = '\alpha''';
    if cfg.plotmodel,
        hold on; plot(t, XM(2, :), 'b'); leg{end+1} = '\alpha''_m';
    end;
    legend(leg{:});
    subplot(413);
    stairs(t(1:end-1), U, 'k');
    legend('u');
    subplot(414); leg = {};
    plot(t, R, 'k'); leg{end+1} = 'r';
    if cfg.plotmodel,
        hold on; plot(t, RM, 'b'); leg{end+1} = 'r_m';
    end;
    legend(leg{:});
end;

% return data in a format compatible with plothistory
trajhist.t = t;
trajhist.x = XN;
trajhist.u = U; trajhist.u(:, end+1) = NaN; % pad with an extra control action at k+1, which is undefined
trajhist.r = R;
% these are extra fields w.r.t. standard
trajhist.X = X;
trajhist.XM = XM;
trajhist.DATA = DATA;
% trajhist = varstostruct('X', 'U', 'R', 'XN', 'XM', 'DATA');

% save data if requested
if ~isempty(cfg.savedatafile),
    save([cfg.savedatadir cfg.savedatafile]);
end;

end         % IPSETUP_CONTROL returning HH, FIGH

function xn = normalizex(x, maxomega)
% State normalization. Brings the angle into [-pi, pi), shifting it
% so that the unstable equilibrium is in 0, and bounds the angular velocity
% Works for scalars or vectors
xn = [mod(x(1, :), 2*pi) - pi; sat(x(2, :), -maxomega, maxomega)];
end