function varargout = ipsetup_lspionline(cfg)
% Online, (semi-)optimistic LSPI for the inverted pendulum.
%   VARARGOUT = LSPI(CFG)
% Online, semi- or fully-optimistic least-squares policy iteration for the inverted pendulum.
% This version works on the real system and takes care of sample timings etc.
% Policy improvements are performed in-between trials, rather than after an arbitrary number
% of samples as in the original LSPIONLINE algorithm.
% See LSPIONLINE for more details.
%
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
%

% Author: Lucian Busoniu
% Version: 1.0

if nargin < 1, cfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
% function config
CFG.run = 0;                        % run learning
CFG.resume = 0;                     % resume learning
CFG.replay = 0;                     % replay learned policy
CFG.sol = 0;                        % plot solution and convergence stats
CFG.evol = 0;                       % plot evolution statistics
CFG.traj = 0;                       % plot some trajectories
CFG.datafile = 'lspios_data';       % save data to file
CFG.datadir = [];                   % data dir
% main algorithm config
CFG.problem = 'ipsetup_problem';    % should never be changed
CFG.approx = [];                    % approximator object (or config for approx)
CFG.gamma = 0.98;                   % discount factor
CFG.maxtime = 5*60;                 % maximum time to run; must be multiple of Ts
CFG.trialtime = 1.5;                % trial length; must be multiple of Ts
CFG.invdelta = 0.001;               % use for inverse computation (has to be >0 since LSPI works online)
CFG.eps = 0;
CFG.reset = 'down';                 % trial reset, one of: down, up, uprand, upranddownrand, randdown
% updates config
CFG.dupd = 1.5;                     % update theta once dupd seconds; must be multiple of Ts
% exploration
CFG.explor = 1;                     % start with this exploration rate
CFG.explordecay = 0.99;             % exploration (exponential) decay rate per second
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 1.5;                     % end time for replay
% stats config
CFG.dstoretheta = 0;                % save theta at most once each dstoretheta seconds (rarer if theta updates rarer)
% display config
CFG.showsamples = 0;                % in the solution plot, put the samples to get an idea of their distrib.
CFG.dtraj = 20;                     % default time interval between trajectories
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.ddisp = 1.5;                    % feedback at most once dsave seconds (rarer if theta updates rarer)
CFG.dsave = 30;                     % save at most once dsave seconds (rarer if theta updates rarer)
CFG.silent = 0;                     % suppress all output
% figure config
CFG = setfigprop(CFG, 'addfields');
CFG.miscinfo = '';                  % field for miscellaneous information

% Early defaults (initialized before calling problem defaults)
ECFG.model_params = {};             % parameters for problem calling in 'model' mode
ECFG.lspi_params = {};              % parameters for problem calling in 'lspi' mode

% List of fields that define the problem
KEEPFIELDS = {'gamma', 'approx'};

% ==== PARSE AND PROCESS CONFIG ====
cfg = parseconfig(cfg, CFG, ECFG, 'lspi');
% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       
% determine if initialization needed; if not, need to load: check if datafile exists
cfg.init = ~(cfg.resume || cfg.replay || cfg.sol || cfg.evol || cfg.traj);
if ~cfg.init && ~exist([cfg.datafile '.mat'], 'file'),
    error(['File [' cfg.datafile '.mat] does not exist. Terminating.']);
end;

% ==== IF NOT INITIALIZING: NEED TO LOAD DATA ====
if ~cfg.init,        % load data file, making sure that cfg and KEEPFIELDS is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
    if exist('htheta', 'var'), 
        error('The algorithm used to produce this data was not LSPI, but LSPIH. Please use LSPIH.');
    end;
    approx = revise_approx(approx);     % make sure approx is in latest format
    cfg.approx = revise_approx(cfg.approx);     % make sure approx is in latest format
end;

% Echo config
dispx('LSPI will run with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

% ==== IF INITIALIZING: CREATE MODEL, AND IF NEEDED: APPROXIMATOR ====
if cfg.init,
    model = ipsetup_problem('model', cfg.model_params{:});
    % heuristic check if we already have an approx object
    if isstruct(cfg.approx) && isfield(cfg.approx, 'N') && isfield(cfg.approx, 'phi') ...
            && isfield(cfg.approx, 'q') && isfield(cfg.approx, 'h'),
            % nothing to do
            dispx('Approx object supplied', cfg.verb, 3);
    else    % create approximator using cfg.approx as the config
        cfg.approx = create_approx(model, cfg.approx);
    end;
end;

% ==== RUN LSPI ====
if cfg.run || cfg.resume,
       
    % shorthand variables
    approx = cfg.approx;
    N = approx.N;
    p = model.p; q = model.q; Ts = model.Ts;
    maxx = model.maxx; maxu = model.maxu;  maxomega = maxx(2);
    % total and trial length
    if mod(cfg.maxtime, Ts) ...
            || (isfinite(cfg.trialtime) && mod(cfg.trialtime, Ts)) ...
            || mod(cfg.dupd, Ts)
        error('maxtime, trialtime, and dupd must be multiples of the sample time');
    end;
    if (cfg.dupd ~= cfg.trialtime),
        disp('For reasons of synchronizing with real-time, trial length must equal dupd');
    end;
    K = cfg.maxtime / Ts;
    if cfg.trialtime < Inf,     Ktrial = cfg.trialtime / Ts; Ntrials = ceil(K / Ktrial);
    else                        Ktrial = K; Ntrials = 1;
    end;
    Kupd = cfg.dupd / model.Ts;     % number of steps between theta updates
    Ksecond = round(1 / model.Ts);  % approximate(!!) second length in # of steps    
    Nupdates = ceil(K / Kupd);      % maximum possible # of updates of theta
    
    % Initialize if not resuming
    if cfg.init,
        % init A and b (never using sparse A since stuff keeps accumulating in it anyway)
        if cfg.invdelta > 0,    A = eye(N, N) .* cfg.invdelta;
        else                    A = zeros(N, N);
        end;
        b = zeros(N, 1);
        % initialize parameter vector to zeroes; old parameter is the same
        theta = zeros(N, 1); oldtheta = theta;

        % Init history 
        % NOTE this init is inappropriate for episodic tasks
        % In such tasks, a trial might have less than Ktrial steps, and there might be
        % more than Ntrials trials needed to accumulate K samples
        X = nan(p, Ktrial+1, Ntrials);
        XS = nan(p, Ktrial+1, Ntrials);     % state of the setup
        U = nan(q, Ktrial+1, Ntrials);
        R = nan(1, Ktrial+1, Ntrials);
        if cfg.dstoretheta > 0,
            % thetah{end} corresponds to the final update of theta
            thetah = {theta};       % for compatibility with old LSPI style
            thetahk = 0;            % theta corresponds to before the 1st sample
        end;
        % first element of this vector will always be NaN, so that elements of these vectors
        % correspond index-wise to the stored values of theta
        % updatesk(end), deltah(end) correspond to the final update of theta
        updatesk = nan(Nupdates+2, 1);  % stores index of sample before update
        deltah = nan(Nupdates+2, 1);

        % init counters, indices
        k = 1;               % sample index
        ktr = NaN;           % sample index within trial (initialized below)
        kupd = 1;            % sample index after last update
        ntr = 0;             % index of current trial
        nupd = 1;            % index of current update
        trun = 0;            % execution time

        explor = cfg.explor;
        % prepare helper variables for discrete-action exploration
        disc = any(strcmp(cfg.approx.type, {'rbfdisc', 'triang'}));
        if disc,        % initialize action grids and their size
            discU = cfg.approx.U;
            discUn = zeros(length(discU)); for i=1:length(discU), discUn(i) = length(discU{i}); end;        
            flatU = flat(discU); nU = size(flatU, 2);
        end;
    end;        % initialization IF

    dispx('Performing online LSPI...', cfg.verb, 0);
    
    conv = 0; tmark = cputime;
    H = ipsetup_reset;          % reset the setup / sensor
    % trials loop
    while ~conv && ntr < Ntrials,
        
        % ==== Initialize trial ====
        % reset the initial state
        xs = localreset(cfg, cfg.reset, H, Ts, maxu, maxomega, ntr);
        % compute normalized initial state; init "old state"
        xsold = xs; x = normalizex(xs, maxomega); 

        ntr = ntr + 1;                  % increment index of current trial        
        ktr = 1;                        % reset step counter
        
        % ==== Samples (steps) loop ====
        tic;
        while ktr <= Ktrial,
            % START CRITICAL 
            % up until END CRITICAL the code needs to be VERY FAST 
            % (and by that I mean MAKE IT SCREAM :) )
            % read sensor data for current sample
            DATA = mops('Read', H);             % read data
            xs(1) = DATA(3);                    % pick up the angle
            xs(2) = (xs(1) - xsold(1)) / Ts; 	% velocity = dangle/dt
            x(1) = mod(xs(1), 2*pi) - pi;       % normalize angle
            x(2) = min(max(xs(2), -maxomega), maxomega);    % bound velocity
            % choose control action
            if rand > explor,   % exploit using current greedy policy
                u = approx.h(approx, theta, x);
            else 
                if disc,        % flatU is a set of discrete actions, each action on a column
                    u = flatU(:, unidrnd(nU));
                else            % U is continuous, assumed symmetrical w/ bounds in maxu
                    u = (2*rand(q, 1) - 1) .* maxu;
                end;
            end;
            % apply control action
            mops('Write', H, 0, 1, u, 0);   % send control input to process
            % END CRITICAL
            
            % starting here, we have some leeway but we still need to be done by the end
            % of the sample time
            
            % save xs, x, u on the histories
            xsold = xs;
            XS(:, ktr, ntr) = xs; X(:, ktr, ntr) = x; U(:, ktr, ntr) = u;

            % process last transition (unless this is the first step of a trial, in which case
            % there is no transition to process)
            if ktr > 1,                
                % compute reward
                R(1, ktr, ntr) = ipsetup_reward(model, X(:, ktr-1, ntr), U(:, ktr-1, ntr), X(:, ktr, ntr));
                
                % greedy action in current state (maybe different from chosen)
                up = approx.h(approx, theta, X(:, ktr, ntr));
                % compute phi(x, u) (time index ktr-1) and phi(x', h(x')) (time index ktr)
                phixu = approx.phi(approx, X(:, ktr-1, ntr), U(:, ktr-1, ntr));
                phixpup = approx.phi(approx, X(:, ktr, ntr), up);

                % update A and b
                A = A + phixu * (phixu - cfg.gamma * phixpup)';
                b = b + phixu * R(:, ktr, ntr);
            end;
            
            % decay exploration (approximately) once every second
            if ~mod(k, Ksecond), explor = explor * cfg.explordecay; end;
            
            % increment sample counters
            k = k + 1; 
            kupd = kupd + 1;
            ktr = ktr + 1;
            
            % wait out the rest of the sample
            while toc < Ts, end; 
            tic;
        end;

        % ==== Trial finished ====

        % Update of A and b with the final transition in the trial
        % read sensor data for last sample
        DATA = mops('Read', H);
        XS(1, ktr, ntr) = DATA(3);
        XS(2, ktr, ntr) = (XS(1, ktr, ntr) - xsold(1)) / Ts;
        X(:, ktr, ntr) = normalizex(XS(:, ktr, ntr), maxomega);
        % process final transition
        % compute reward
        R(1, ktr, ntr) = ipsetup_reward(model, X(:, ktr-1, ntr), U(:, ktr-1, ntr), X(:, ktr, ntr));
        % greedy action in last state
        up = approx.h(approx, theta, X(:, ktr, ntr));
        % compute phi(x, u) (time index ktr-1) and phi(x', h(x')) (time index ktr)
        phixu = approx.phi(approx, X(:, ktr-1, ntr), U(:, ktr-1, ntr));
        phixpup = approx.phi(approx, X(:, ktr, ntr), up);
        % update A and b
        A = A + phixu * (phixu - cfg.gamma * phixpup)';
        b = b + phixu * R(:, ktr, ntr);
        
        % REMARK this is not be necessary given that we control below
%         % === Give a push so weight doesn't sit upwards due to static friction ===
%         mops('Write',H,0,1,2,0); tic; while toc < 0.2, end; mops('Write',H,0,1,0,0);
        
        % === While the weight is falling, do a policy improvement ===
        theta = A \ b;
        deltah(nupd+1) = norm(theta - oldtheta, 2);
        oldtheta = theta;
        % check convergence condition
        conv = deltah(nupd+1) <= cfg.eps;
        % store new parameter vector if required and passed time boundary
        if cfg.dstoretheta > 0, 
            if (nupd == 1 && floor(k*Ts / cfg.dstoretheta) > 0) ...
                || (nupd > 1 && floor(k*Ts / cfg.dstoretheta) > floor(updatesk(nupd-1)*Ts / cfg.dstoretheta)),
                % save a theta snapshot
                thetah{end+1} = theta; 
                % remember after which sample we took the snapshot
                thetahk(end+1) = k - 1;
            end;
        end;
        % remember after which sample we updated
        updatesk(nupd) = k - 1;     
        % console feedback if we passed a ddisp boundary
        if (nupd == 1 && floor(k*Ts / cfg.ddisp) > 0) ...
            || (nupd > 1 && floor(k*Ts / cfg.ddisp) > floor(updatesk(nupd-1)*Ts / cfg.ddisp)),
            dispx(['k=' num2str(updatesk(nupd)) ' (t=' num2str(updatesk(nupd) * Ts) ...
                '), updated #' num2str(nupd) '. delta=' num2str(deltah(nupd+1))], cfg.verb, 3);
        end;
        % data backup if we passed a dsave boundary
        if (nupd == 1 && floor(k*Ts / cfg.dsave) > 0) ...
            || (nupd > 1 && floor(k*Ts / cfg.dsave) > floor(updatesk(nupd-1)*Ts / cfg.dsave)),
            trun = trun + cputime - tmark; tmark = cputime;
            save(cfg.datafile);
            dispx(['Data at k=' num2str(updatesk(nupd)) ' (t=' num2str(updatesk(nupd) * Ts) ...
                ') saved to [' cfg.datafile '].'], cfg.verb, 1);
        end;
        
        nupd = nupd + 1;            % update counter of updates
        kupd = 1;                   % reset sample index after last update
    end;        % trial WHILE loop

    % after the experiment is done, stabilize pointing down
	stabilize(H, maxu, Ts);
    % stop motor -- reset actuators, close the port
    mops('Write', H, 0, 1, 0, 0.0);
    mops('CloseAll');
    
    clear oldtheta;

    if conv,        dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    else            dispx(['maxtime=' num2str(cfg.maxtime) ' exhausted. Algorithm stopped'], cfg.verb, 0);
    end;
    
end;    % end of RUN branch


% Options below might create figures, init figure handles array
figh = [];

% ==== REPLAY POLICY ====
if cfg.replay,

    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    xs = zeros(model.p, length(t)); 
    x = xs; 
    u = zeros(1, length(t)); u(end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % reset driver
    H = ipsetup_reset;
    % reset system to initial state
    xs0 = localreset(cfg, cfg.x0, H, Ts, maxu, maxomega, []);   % [] = replay reset, no trial index
    
    % put the initial state on the setup state vector
    xs(:, 1) = xs0;    
    % steps loop
    tic;
    for k = 1:Ns,
        DATA = mops('Read', H);        % read sensor data
        xs(1, k) = DATA(3);
        if k > 1,  xs(2, k) = (DATA(3) - xs(1, k-1)) / Ts; end;
        x(:, k) = normalizex(xs(:, k), maxomega);
        u(k) = approx.h(approx, theta, x(:, k));
        mops('Write', H, 0, 1, u(k), 0);    % send control input to process
        if k > 1, r(k) = ipsetup_reward(model, x(:, k-1), u(k-1), x(:, k)); end;
        while toc < Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    % pick up last sample
    xs(1, k+1) = DATA(3);
    xs(2, k+1) = (DATA(3) - xs(1, k)) / Ts;
    x(:, k+1) = normalizex(xs(:, k+1), maxomega);
    % stabilize downward
	stabilize(H, maxu, Ts);
    % stop motor -- reset actuators, close the port
    mops('Write', H, 0, 1, 0, 0.0);
    mops('CloseAll');
    
    % plot history & optionally save figures
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);
    hist.R = discreturn(cfg, hist.r, Ns, 0);    % 0 = not terminal
    if ~cfg.silent,
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        else
            figh(end+1) = plothistory(hist);
        end;
        setfigprop(cfg);
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay


% ==== STATISTICS ====
if (cfg.evol || cfg.sol || cfg.traj) && ~cfg.silent,
    % grayscale styles
    gs.trajsep = [.5 .5 .5];
    gs.cm = gray(96); gs.cm = gs.cm(24:end);
    % color styles
    cs = gs;
    cs.trajsep = 'r';
    cs.cm = jet(64);
    % set style
    if cfg.grayscale, sty = gs; else sty = cs; end;    
    
    % readable labels
    labels.time = 't [s]'; 
    labels.x = {'\alpha [rad]', '\alpha'' [rad/s]'};
    labels.u = {'u [V]'};
    labels.r = 'r [-]';
    labels.h = 'h(x) [V]'; labels.V = 'V(x)'; 
    labels.delta = {'$\Vert \theta_\ell - \theta_{\ell-1}\Vert_2$', 'Interpreter', 'Latex', 'FontSize', 14}; 
    labels.condA = 'cond(A)';
end;

if cfg.sol && ~cfg.silent,
    figh(end+1) = figurex([1100 450]); % colormap(sty.cm);
    K = find(~isnan(deltah), 1, 'last') - 1;
    subplot(3, 2, [1 3]); cla;
    approx.plotv(approx, theta); title(labels.V);
    subplot(3, 2, [2 4 6]); cla;
    approx.ploth(approx, theta, 'npoints=100'); title(labels.h);
    subplot(3, 2, 5); cla;
    % -1 to get rid of the extra sample due to the "cursor" nature of the sample indices
    semilogy(updatesk(2:K+1) * Ts, deltah(2:K+1), 'k'); grid on; xlabel(labels.time); ylabel(labels.delta{:});
    % show samples distr if requested
    if cfg.showsamples,
        disp('Press any key to show x_k samples...');
        pause;
        x1 = reshape(X(1, :, :), 1, []);
        x2 = reshape(X(2, :, :), 1, []);
        valid = ~isnan(x1) & ~isnan(x2);
        x1 = x1(valid); x2 = x2(valid);
        subplot(3, 2, [1 3]); hold on;
        z = get(gca, 'zlim'); z = z(1) + 0 * x1;
        plot3(x1, x2, z, 'LineStyle', 'none', 'MarkerEdgeColor', 'none', 'MarkerFaceColor', 'r', ...
            'Marker', 'o', 'MarkerSize', 2);
        subplot(3, 2, [2 4 6]); hold on;
        plot(x1, x2, 'LineStyle', 'none', 'MarkerEdgeColor', 'none', 'MarkerFaceColor', 'r', ...
            'Marker', 'o', 'MarkerSize', 2);
    end;
    setfigprop(cfg);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% Plot some trajectories
% Note that we use specific code for the IP, unlike the general lspionline
% For instance, reward plot is half the height of the rest
if cfg.traj && ~cfg.silent,
    figh(end+1) = figure;
    ntraj = 1;
    % k'screen' refers to coordinates on screen (time lapse)
    % k, trajk refers to (simulated) real-time coordinates
    k = 0; kscreen = 0; trajk = []; trajkscreen = []; xticklabels = {};
    for i = 1:size(X, 3),
        if isscalar(cfg.dtraj),     plotit = k*Ts >= (ntraj-1)*cfg.dtraj;
        else                        plotit = k*Ts >= cfg.dtraj(ntraj);
        end;
        trajlen = find(~isnan(U(1, :, i)), 1, 'last');  % not counting the time step AFTER the last control
        if plotit,
            for ip = 1:p,
                subplot(2*(p+q)+1, 1, 2*(ip-1) + [1 2]); hold on;
                plot((kscreen + (0:trajlen))*Ts, X(ip, 1:trajlen+1, i), 'k-');
            end;
            for iq = 1:q,
                subplot(2*(p+q)+1, 1, 2*(p+iq-1) + [1 2]); hold on;
                plot((kscreen + (0:trajlen-1))*Ts, U(iq, 1:trajlen, i), 'k-');
            end;
            subplot(2*(p+q)+1, 1, 2*(p+q)+1); hold on;
            plot((kscreen + (1:trajlen))*Ts, R(1, 2:trajlen+1, i), 'k-','LineWidth',1,'Color',[.6,.6,.6]);
            trajkscreen(end+1) = kscreen;
            trajk(end+1) = k;
            ntraj = ntraj + 1;
            xticklabels{end+1} = sprintf('%d', k * Ts);
            kscreen = kscreen + trajlen + 1;    % counting the extra samples after the last controls
        end;
        k = k + trajlen;
    end;
    % finally add some markers in-between, and zero-lines
    for ip = 1:p+q+1,
        if ip == p+q+1, % last, reward plot, stretching just 1 subplot
            subplot(2*(p+q)+1, 1, 2*(ip-1)+1);
        else
            subplot(2*(p+q)+1, 1, 2*(ip-1) + [1 2]);
        end;
        y = get(gca, 'YLim');
        for i = 1:length(trajk), line(trajkscreen(i)*Ts + [0 0], y, 'Color', sty.trajsep); end;
        if ip <= p, 
            xlim([0 kscreen] * Ts);
            ylabel(labels.x{ip}); 
            set(gca, 'XTick', trajkscreen * Ts);
            set(gca, 'XTickLabel', xticklabels);
        elseif ip <= p+q, 
            xlim([0 kscreen] * Ts);
            ylabel(labels.u{ip-p}); 
            set(gca, 'XTick', trajkscreen * Ts);
            set(gca, 'XTickLabel', xticklabels);
        else % last plot, also add time lables
            xlim([0 kscreen] * Ts);
            ylabel(labels.r); xlabel(labels.time);
            set(gca, 'XTick', trajkscreen * Ts);
            set(gca, 'XTickLabel', xticklabels);
        end;
        % if 0 within axes limits, plot a zero-line
        if y(1) < 0 && y(2) > 0,
            x = get(gca, 'XLim');
            plot(x, [0 0], 'k:');
        end;
    end;
    set(gcf, 'NumberTitle', 'off', 'Name', cfg.datafile);
    setfigprop(cfg);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% ==== FINAL SAVE DATA ====
if cfg.run || cfg.resume, 
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which('ipsetup_problem'));
    else
        datadir = cfg.datadir; 
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        if exist([CFG.datafile '.mat'], 'file'), delete([CFG.datafile '.mat']); end;
    elseif ~strcmp(datadir, pwd)
        delete([cfg.datafile '.mat']);   % need to save in a different directory, so delete anyway
    end;    % otherwise just overwrite
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% set output
if cfg.run || cfg.resume,       % output optimal solution
    varargout = {theta};
elseif cfg.replay,              % output history and possibly figure handles
    varargout = {hist, figh};
elseif cfg.evol || cfg.sol || cfg.traj,               % output fig handles (replay takes precedence)
    varargout = {figh};
end;

end
% lspi() RETURNING varargout =================================================================

% ========================
% === LOCAL FUNCTIONS ====
% ========================

function xs = localreset(cfg, restype, H, Ts, maxu, maxomega, ntr)
% Local function to reset the pendulum initial state
%   restype -- type of the reset
%   ntr     -- index of current trial; if empty this is a reset for a replay run,
%           and there are no trials

% the modes in the outer switch work at trial start or for replay
switch restype,
    case 'rand',        % apply some random actions
        [H, xs] = randomcontrol(H, maxu + 2, Ts, 1, 30); % for 1 second, lasting 30 steps each
    case 'down',        % pointing down
        [H, xs] = stabilize(H, maxu, Ts);   
    case 'up',          % pointing up
        [H, xs] = swingup(H, maxu, maxomega, Ts);
    case 'uprand',      % point upwards then apply some random actions
        [H, xs] = swingup(H, maxu, maxomega, Ts);
        [H, xs] = randomcontrol(H, maxu, Ts, 0.2, 10);
    otherwise,
        % the rest of the possible reset modes only work at trial start, during learning
        if ~isempty(ntr) && ntr >= 0,
            switch restype,
                case 'upranddownrand',% half upward, half downward; + some random actions
                    if mod(ntr, 2),
                        [H, xs] = swingup(H, maxu, maxomega, Ts);
                        [H, xs] = randomcontrol(H, maxu, Ts, 0.2, 10);
                    else
                        [H, xs] = stabilize(H, maxu, Ts);
                        [H, xs] = randomcontrol(H, maxu, Ts, 0.5, 20);
                    end;
                case 'randdown',    % half random, half downward
                    if ~mod(ntr, 2),    % point down
                        [H, xs] = stabilize(H, maxu, Ts);
                        dispx(['Resetting [' restype ']; now down.'], cfg.verb, 4);
                    else                % apply random actions
                        % for 1 second, lasting 30 steps each
%                         [H, xs] = randomcontrol(H, maxu + 2, Ts, 1, 30); 
                        [H, xs] = randomcontrol(H, maxu + 2, Ts, 1, 10); 
                        dispx(['Resetting [' restype ']; now rand.'], cfg.verb, 4);
                    end;
                case 'downranddown',    % half down+rand, half down
                    if ~mod(ntr, 2),    % point down
                        [H, xs] = stabilize(H, maxu, Ts);
                        dispx(['Resetting [' restype ']; now down.'], cfg.verb, 4);
                    else                % down, then apply random actions
                        [H, xs] = stabilize(H, maxu, Ts);
                        [H, xs] = randomcontrol(H, maxu + 2, Ts, 1, 30); 
                        dispx(['Resetting [' restype ']; now down+rand.'], cfg.verb, 4);
                    end;
                otherwise,
                    dispx(['Warning! Unknown reset method [' restype ']. Doing nothing.'], cfg.verb, 2);
            end;
        else 
            % not at trial start (or unkown method)
            dispx(['Warning! Reset method only [' restype '] working at trial start, or is unknown.'], cfg.verb, 2);
        end;
end;
end     % END ipslspio_reset() returning xs, the final state


function xn = normalizex(x, maxomega)
% State normalization. Brings the angle into [-pi, pi), shifting it
% so that the unstable equilibrium is in 0, and bounds the angular velocity
% Works for scalars or vectors
xn = [mod(x(1, :), 2*pi) - pi; min(max(x(2, :), -maxomega), maxomega)];
end         % normalizex()

% ==== Local functions for various control laws
function [H, xs] = swingup(H, maxu, maxomega, Ts)
% swing the weight up using a heuristic policy (which btw works quite well -- thanks to
% Maarten)
    kk = 1; XX = NaN(2, ceil(1.5/Ts));    % provide for 1 second (should be enough to stabilize)
    % control until speed becomes (nearly) 0 rad/sec (max limit at 5 seconds)
    stable = 0;
    tic; 	% reset Matlab's tic-toc timer
    while ~stable && kk < 1.5/Ts,
        DATA = mops('Read', H);        % read sensor data
        % current angle
        XX(1, kk) = DATA(3);
        % velocity = dangle/dt
        if kk > 2,  XX(2, kk) = (DATA(3) - XX(1, kk-1)) / Ts;
        else        XX(2, kk) = DATA(3) / Ts;
        end;
        x = normalizex(XX(:, kk), maxomega);
        if abs(x(1)) > pi/2,    u = maxu * (sign(x(2)) >= 0);
        else                    u = -10 * x(1) -1 * x(2);
        end;
        mops('Write', H, 0, 1, sat(u, -maxu, maxu), 0);  % send saturated control input to process
        stable = kk > 20 && all(abs(XX(2, kk-20:kk)) < 0.001);
        kk = kk + 1;
        while toc < Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    mops('Write', H, 0, 1, 0, 0);   % send control input to process    
    xs = XX(:, kk - 1);
end         % swingup()

function [H, xs] = stabilize(H, maxu, Ts)
% Stabilize pointing downward
%     disp('Stabilizing...');
    kk = 1; XX = NaN(2, ceil(1/Ts));    % provide for 1 second (should be enough to stabilize)
    % control until speed becomes (nearly) 0 rad/sec (max limit at 5 seconds)
    stable = 0;
    tic; 	% reset Matlab's tic-toc timer
    while ~stable && kk < 3/Ts,
        DATA = mops('Read', H);        % read sensor data
        % current angle
        XX(1, kk) = DATA(3);
        % velocity = dangle/dt
        if kk > 2,  XX(2, kk) = (DATA(3) - XX(1, kk-1)) / Ts;
        else        XX(2, kk) = DATA(3) / Ts;
        end;
        % reference (stable eqn) = 0
        % Wrap the angle so we don't care about # full rotations
        err = 0 - (mod(XX(1, kk) + pi, 2*pi) - pi); 
        if kk > 2,  derr = (err - (0 - (mod(XX(1, kk-1) + pi, 2*pi) - pi))) / Ts;
        else        derr = err / Ts;
        end;
        u = 25 * err + 1 * derr;         % PD control law
        mops('Write', H, 0, 1, sat(u, -maxu, maxu), 0);  % send saturated control input to process
        stable = kk > 20 && all(abs(XX(2, kk-20:kk)) < 0.001);
        kk = kk + 1;
        while toc < Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    mops('Write', H, 0, 1, 0, 0);   % send control input to process    
    H = ipsetup_reset;          % reset the setup / sensor
    xs = [0; 0];
%     plot(XX'); disp([8 'done.']); pause;
end         % stabilize()

function [H, xs] = randomcontrol(H, maxu, Ts, T, keep)
% apply some random control actions
    oldx = [0; 0]; x = 0*oldx; tic;
    for kk = 1:T/Ts,                            
        data = mops('Read',H);              % read sensor data
        x(1) = data(3); x(2) = (data(3) - oldx(1)) / Ts; 
        oldx = x;
        if ~mod(kk-1, keep), u = (2*rand - 1) * maxu; end;
        u = sat(u, -maxu, maxu);            % saturate control action
        mops('Write', H, 0, 1, u, 0);       % send control input to process
        while toc < Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    xs = x;
	mops('Write', H, 0, 1, 0, 0);   % send control input to process    
end         % randomcontrol()
