function out = rotarypend_problem(what, varargin)
% Integrator problem setup.
%   MODEL = RESONATINGARM_PROBLEM(MODE, [CONFIG])
% This function conforms to the specifications established by SAMPLE_PROBLEM.
% Second argument for 'model' mode is the model config.

switch what
    case 'shortid', 
        out = 'rotpend';
    
    case 'model'
        % #### Standard fields
        CFG.id = 'rotpend';
        CFG.p = 4; CFG.q = 1;
        CFG.maxx = [pi; 100; pi; 100];
        CFG.maxu = 9;
        CFG.det = 1;
        CFG.userefmodel = 0;
        CFG.fun = @rotarypend_mdp;
        CFG.plotfun = @rotarypend_plot;
        
        % #### Physical parameters
        CFG.Beq = 0.004;      % Equivalent viscous damping coefficient
        CFG.Jeq = 0.0035842;  % Moment of inertia of the arm and pendulum about the axis of ?
        CFG.Jm = 3.87e-7;     % Moment of inertia of the rotor of the motor
        CFG.Kg = 70;          % SRV02 sys. gear ratio (14*5)
        CFG.Km = 0.00767;     % Back-emf const.
        CFG.Kt = 0.00767;     % Motor-torque const.
        CFG.L = 0.1685;       % half length of pendulum
        CFG.m = 0.127;        % mass of pendulum
        CFG.r = 0.216;        % rotating arm length
        CFG.Rm = 2.6;         % armature resistance
        CFG.NUg = 0.9;        % gearbox efficiency
        CFG.NUm = 0.69;       % motor efficiency
        % #### Numerical integration
        CFG.Ts = 0.05;              % scalar, sampling time      
        CFG.odemethod = 'euler';    % 'varode', 'fixedode', 'eul'
        % CFG.odesolver = @ode4;    % solver function
        CFG.odesteps = 10;          % # of steps per sample time (ignored for varode)

        % #### Reward config
        CFG.rewtype = 'nlqr';       % 'expref'onential or 'nlqrref'
        % CFG.Qrew = diag([0.1 0.001 1 0.01]); CFG.Rrew = 0.07;
        CFG.Qrew = diag([0 0 1 0]); CFG.Rrew = 0;
        
        % ######## Process the config into the model structure
        if nargin == 1 || isempty(varargin{1}), m = CFG;
        else                                    m = parseconfig(varargin{1}, CFG);
        end;
        
        % physics derived params
        m.a = m.Jeq + m.m * m.r^2 + m.NUg * m.Kg^2 * m.Jm;
        m.b = m.m * m.L * m.r;
        m.c = 4/3 * m.m * m.L^2;
        m.d = m.m * 9.80665 * m.L;
        m.e = m.Beq + (m.NUm * m.NUg * m.Kt * m.Kg^2 * m.Km) / m.Rm;
        m.f = (m.NUm * m.NUg * m.Kt * m.Kg) / m.Rm;
        
        % reward derived params 
        switch m.rewtype,
            case 'nlqr',     % LQR normalized to interval [0, 1]
                m.nonnorm_maxr = m.maxx' * m.Qrew * m.maxx + m.Rrew * m.maxu^2;
                m.maxr = 1;
        end;
        % num int: process derived params 
        switch(m.odemethod)
            case 'varode',
                m.odet = [0 m.Ts/2 m.Ts];
                m.odeopt = odeset;
            case 'fixedode',
                m.odet = 0 : m.Ts / m.odesteps : m.Ts;
        end;
        
        out=m;
        
    case 'plotfun', 
        out = @rotarypend_plot;
      
    case 'reset',   % handle state reset, only single-value supported
        % model = varargin{1};
        % res = varargin{2};                 
        out = varargin{2};
        
end;

end % END doubleint_problem(), RETURNING out ====================================


