% Script to test various extra-trees implementations

%% Initialize
clear all; close all;
Nruns = 1;          % how many times to run tree creation & query (to get average run times)

% % small test for visualization
% maxx = [pi 10*pi]'; % dimension of (symmetrical) input space
% D = length(maxx);   % dimensionality
% Ns = 500;           % number of samples to build the tree from
% Nq = 10;            % number of query samples
% K = 20;

% large test for execution time or extensive error checks
maxx = [pi 10*pi pi 15*pi 10 30]';  % dimension of (symmetrical) input space
D = length(maxx);   % dimensionality
Ns = 5000;         % number of samples to build the tree from
Nq = 1000;         % number of query samples
Ntrees = 50;
ktrees = length(D);
nmintrees = 2;
fun = @(x) x(1)^2 + 0.1 * x(2) - x(3)^2 + exp(-abs(x(4))) + x(5)/10 - x(6)/30;

Xs = rand(D, Ns) .* repmat(maxx, 1, Ns);
Ys = zeros(1, Ns); for is = 1:Ns, Ys(is) = fun(Xs(:, is)); end;
Xq = rand(D, Nq) .* repmat(maxx, 1, Nq);


%% Test CPP kd-tree implementation -- by Pierre Geurts
tcpp = struct;
rcpp = struct; rcpp.name = 'CPP extra-trees'; rcpp.Yq = nan(1, Nq);

tcpp.initruns = zeros(Nruns, 1);
tcpp.queryruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    
    % create tree
    disp(sprintf('%s/%d: %s', rcpp.name, i, 'creating ensemble'));
    etcfg = init_extra_trees(ktrees);
    etcfg.nbterms = Ntrees;
    etcfg.rtparam.nmin = nmintrees;
    [qdummy, dummy, ens] = rtenslearn_c(single(Xs'), single(Ys'), int32(1:Ns), [], etcfg, single(Xq(:, 1)'), 0);
    clear qdummy dummy
    
    tcpp.initruns(i) = cputime - tmark;
    tmark = cputime;
    
    disp(sprintf('%s/%d: %s', rcpp.name, i, 'querying ensemble'));
    % iterative variant
%     for iq = 1:Nq, Yq(iq) = rtenspred(ens, single(Xq(:, iq)')); end;
    % single-call variant
    Yq = rtenspred(ens, single(Xq'));     
    % NOTE! querying N times takes N times as long ==> vectorization is key
    
    tcpp.queryruns(i) = cputime - tmark;
end;

tcpp.totalruns = tcpp.initruns + tcpp.queryruns;
tcpp.init = mean(tcpp.initruns);
tcpp.query = mean(tcpp.queryruns);
tcpp.total = tcpp.init + tcpp.query;
tcpp.querypersample = tcpp.query / Nq;


%% Test Matlab implementation --  by Raphael Fonteneau
tmtb = struct;
rmtb = struct; rmtb.name = 'Matlab extra-trees'; rmtb.Yq = nan(1, Nq);

tmtb.initruns = zeros(Nruns, 1);
tmtb.queryruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    
    % create tree
    disp(sprintf('%s/%d: %s', rmtb.name, i, 'creating ensemble'));
    ens = build_an_extra_tree_ensemble([Xs; Ys]',Ntrees,ktrees,nmintrees);
    tmtb.initruns(i) = cputime - tmark;
    tmark = cputime;
    
    disp(sprintf('%s/%d: %s', rmtb.name, i, 'querying ensemble'));
    Yq = zeros(1, Nq);
    for iq = 1:Nq, Yq(iq) = ET(Xq(:, iq)', ens); end;
    
    tmtb.queryruns(i) = cputime - tmark;
end;

tmtb.totalruns = tmtb.initruns + tmtb.queryruns;
tmtb.init = mean(tmtb.initruns);
tmtb.query = mean(tmtb.queryruns);
tmtb.total = tmtb.init + tmtb.query;
tmtb.querypersample = tmtb.query / Nq;



