function  varargout = soop(x0, model, cfg)
% SOOP Planning based on sequential optimistic optimization
%   U0 = SOOP(X0, MODEL, CFG)
%   [U0, PLANINFO] = SOOP(X0, MODEL, CFG)
%   [U0, PLANINFO, PLANDATA] = SOOP(X0, MODEL, CFG)
%   [USEQ, XSEQ, PLANINFO, PLANDATA] = SOOP(X0, MODEL, CFG)
% In the first three modes, just the root action is returned
% In the last two modes, the action sequence and the corresponding state sequence are returned

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % simulation budget
CFG.dimselect = [];             % selection method for which dimension to split:
    % 'discount' for discount-factor-based selection -- the only one supported
CFG.alpha = [];                 % discount factor for dimselect='discount'
CFG.k = -1;                     % discrete time step
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

% options that would maybe make sense, but are not implemented so far
% CFG.refx = [];                  % ref model state at root
% CFG.reusetree = 0;              % try reusing previous tree (only when k is set and > 1)
% CFG.reusetree_maxsize = 0;      % do not reuse tree if larger (has more nodes) than this; 0 = unlimited size

cfg = checkparams(cfg, CFG);

% parameter and error checks
if model.q > 1, error('SOOP currently supports only scalar actions'); end;

% quick-access to config parameters, and other helper vars
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
p = model.p; maxu = model.maxu;

% start counting time
t = cputime;

% init data structures
initlen = 50;                                   % an initial max length guess
initsize = ceil(1.1 * cfg.n);                   % rough upper bound on #seq
SU = NaN(initlen, initsize);                    % (starting points of) U domains
SS = zeros(initlen, initsize);                  % number of splits (could be uint16, with conversion headaches)
SX = NaN(p, initlen+1, initsize);               % states
SR = zeros(initlen, initsize);                  % rewards (note they are not shifted by index 1, as usual)

% start by trisecting the first dimension
for i = 1:3,
    SU(1, i) = (i-1)/3;
    SS(1, i) = 1;
    SX(:, 1, i) = x0; [SX(:, 2, i), SR(1, i)] = model.fun(model, x0, unnormu((i-0.5)/3, maxu));
end;
Ns = 3;             % number of sequences 
n = 3;              % budget consumed so far
longest = 1;        % length of longest sequence

%#### Main loop
stopflag = false; tau = 0; 
while ~stopflag,
    tau = tau + 1;      % this is to make sure tau goes from 1 to n and stops there
   
    %## Select sets/sequences to expand
    expand = false(1, Ns);
    % compute returns of all sequences (note all uninitialized rewards are 0)
    ret = (cfg.gamma .^ (0:longest-1)) * SR(1:longest, 1:Ns);
    for is = 1:Ns,
        % find all sets j >> i (partially greater than i)
        % where j >> i  if  splits j <= # splits i
        greater = all(SS(1:longest, 1:Ns) <= repmat(SS(1:longest, is), 1, Ns), 1);
        % a set is expanded if has a better (or equal) return than all sets partially greater than it
        expand(is)= all(ret(is) >= ret(greater));
        % view ranking if configured
        if cfg.verb >= 5, viewranking(SS, Ns, ret, is, greater, expand); pause; end;
    end;
    
    %## Expand the selected sets/sequences
    for is = find(expand),
        if cfg.verb >= 4, viewseq(SU, SS, 1:Ns, 1 + (tau == 1), is, []); end;   % view with sequence-to-expand in focus

        % rank dimensions by discounted size; note we also rank len+1 dimension (requiring SS=0 for that one)
        len = find(SS(1:longest, is), 1, 'last'); 
        score = (cfg.alpha.^(0:len)) .* (1/3).^SS(1:len+1, is)';
        [maxscore, ksplit] = max(score);
        % "left" and "right" derived sequences inherit info from "middle" sequence
        SU(1:len, Ns+1) = SU(1:len, is); SU(1:len, Ns+2) = SU(1:len, is);
        SS(1:len, Ns+1) = SS(1:len, is); SS(1:len, Ns+2) = SS(1:len, is);
        % transitions depend on full action sequence, so they only get copied up to ksplit-1
        SX(:, 1:ksplit, Ns+1) = SX(:, 1:ksplit, is); SX(:, 1:ksplit, Ns+2) = SX(:, 1:ksplit, is);
        SR(1:ksplit-1, Ns+1) = SR(1:ksplit-1, is); SR(1:ksplit-1, Ns+2) = SR(1:ksplit-1, is);
        JS = [Ns+1 is Ns+2];    % indices of sequences we work with: left, mid, right
        if ksplit == len+1,     
            % trisect the next dimension
            for j = 1:3,
                js = JS(j);
                SU(ksplit, js) = (j-1)/3;
                SS(ksplit, js) = 1;
                [SX(:, ksplit+1, js), SR(ksplit, js)] = model.fun(model, SX(:, ksplit, js), unnormu((j-0.5)/3, maxu));
            end;
            % update resource counters
            n = n + 3; Ns = Ns + 2; 
            % max seq length may have increased
            if ksplit > longest, longest = ksplit; end;
        else
            % trisect an existing dimension
            for j = 1:3,
                js = JS(j);
                % all three sets change starting point, and are split once more
                SU(ksplit, js) = SU(ksplit, js) + (j-1)/3 * (1/3)^SS(ksplit, js);
                SS(ksplit, js) = SS(ksplit, js) + 1;
                % for middle set, the center action samples are the same => so are rewards and states
                if j == 2, continue; end; 
                % for the others, we must simulate
                for k = ksplit:len,
                    [SX(:, k+1, js), SR(k, js)] = model.fun(model, SX(:, k, js), ...
                        unnormu(SU(k, js) + 0.5*(1/3)^SS(k, js), maxu));
                end;
            end;
            % update resource counters
            n = n + 2*(len-ksplit+1); Ns = Ns + 2;        
        end;

        [crtlen, crtsize] = size(SU);
        % if there would be no more room for new sequences, expand the data structures
        % -- with current initialization this should rarely happen
        if Ns+2 > crtsize,
            delta = ceil(0.5*crtsize);  % expand to 150%
            SU(:, crtsize+(1:delta)) = NaN(crtlen, delta); SS(:, crtsize+(1:delta)) = zeros(crtlen, delta);
            SX(:, :, crtsize+(1:delta)) = NaN(p, crtlen+1, delta); SR(:, crtsize+(1:delta)) = zeros(crtlen, delta);
            crtsize = crtsize + delta;
        end;
        % expand length if some set has the last dimension initialized
        % (note we always need an extra dimension with zero splits, for the dimension selection criterion)
        if crtlen == longest,
            delta = ceil(0.5*crtlen);   % expand to 150%
            SU(crtlen+(1:delta), :) = NaN(delta, crtsize); SS(crtlen+(1:delta), :) = zeros(delta, crtsize);
            SX(:, crtlen+1+(1:delta), :) = NaN(p, delta, crtsize); SR(crtlen+(1:delta), :) = zeros(delta, crtsize);
        end;

        if cfg.verb >= 4, viewseq(SU, SS, JS, false, [], JS); pause; end;   % view with expanded seqs in focus
    end;    % end of loop over sequences to expand
    
    % stop when the budget has been exhausted
    stopflag = n >= cfg.n;    
    
    if cfg.verb >= 2,
        fprintf('Iter %d: %d expanded, n=%d trans, #seq=%d, longest=%d\n', ...
            tau, nnz(expand), n, Ns, find(any(SS, 2), 1, 'last')); 
    end;

end;    % end main tree construction loop

%#### Find best action sequence, breaking ties in favor of the first found
ret = (cfg.gamma .^ (0:longest-1)) * SR(1:longest, 1:Ns);
[retstar, istar] = max(ret);
lenstar = find(SS(1:longest, istar), 1, 'last'); 
useq = (unnormu(SU(1:lenstar, istar) + 0.5 .* ((1/3).^SS(1:lenstar, istar)), maxu))';
xseq = SX(:, 1:lenstar+1, istar);
if cfg.verb >= 3, viewseq(SU, SS, 1:Ns, 1, istar, []); pause; end;  % view selected sequence

% create stats if requested on output
if nargout >= 2,
    planinfo.n = n;
    planinfo.maxseqlen = longest;
    planinfo.trun = cputime - t;
    planinfo.nsequences = Ns;
    planinfo.message = sprintf('Planning done, n=%d transitions, len_returned=%d, longest=%d, #seq=%d, trun=%.2f', ...
        planinfo.n, lenstar, planinfo.maxseqlen, planinfo.nsequences, planinfo.trun); 
end;

% create plandata if requested on output
if nargout >= 3,
    plandata.SU = SU(1:longest, 1:Ns);
    plandata.SS = SS(1:longest, 1:Ns);
    plandata.SX = SX(:, 1:longest+1, 1:Ns);
    plandata.SR = SR(1:longest, 1:Ns);
    plandata.useq = useq;
    plandata.xseq = xseq;
end;

% construct the output based on calling mode
switch nargout,
    case 1, varargout = {useq(:, 1)};
    case 2, varargout = {useq(:, 1), planinfo};
    case 3, varargout = {useq(:, 1), planinfo, plandata};
    case 4, varargout = {useq, xseq, planinfo, plandata};
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
% normalize / unnormalize actions
% function u = normu(u, maxu)   % currently not needed
% u = 0.5 + 0.5 * u / maxu;
% end
function u = unnormu(u, maxu)
u = (u - 0.5) * 2 * maxu;
end

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Visualization
% sequences in the U space
function viewseq(SU, SS, indices, reset, focus1, focus2)
longest = size(SS, 1);
if reset >= 2,    
    figurex('h=10 size=[1100,800]'); clf; 
elseif reset >= 1,
    clf;
end;
for is = indices,
   len = find(SS(:, is), 1, 'last'); 
   if any(is == focus1), c = 'r'; offset = -.2;
   elseif any(is == focus2), c = 'b'; offset = 0;
   else c = 'k'; offset = 0;
   end;
   for k = 1:len
       % plot subset
       plot(is-1 + (SU(k, is) + [0 (1/3)^SS(k, is)]), [k k] + offset, c, 'Marker', '.', 'MarkerFaceColor', c);
       hold on;
       scatter(is-1 + SU(k, is) + 0.5*(1/3)^SS(k, is), k + offset, [], c, 'x');
   end;
   plot([is is], [0 longest], 'k:');    % seq separator
end;
end

% ranking of sequences by partial ordering / return
function viewranking(SS, Ns, ret, is, greater, expand)
greater = greater'; greater(is) = false;
figurex('h=20 size=[900,800]');
clf;
offset_x = (max(ret) - min(ret)) * .01;
offset_y = Ns * .01;
% generate labels
labels = cell(Ns, 1);
for i = 1:Ns, 
    len = find(SS(:, i), 1, 'last');
    labels{i} = num2str(SS(1:len, i)');
    labels{i}(labels{i} == ' ') = [];
end;
%
regular = true(Ns, 1);
regular = regular & ~greater; regular(is) = false;
scatter(ret(regular), find(regular), [], 'k', 'MarkerFaceColor', 'k');
hold on;
text(ret(regular) + offset_x, find(regular) + offset_y, {labels{regular}});
%
scatter(ret(greater), find(greater), [], 'b', 'MarkerFaceColor', 'b');
h = text(ret(greater) + offset_x, find(greater) + offset_y, {labels{greater}});
set(h, 'Color', 'b');
%
if expand(is), c = 'r'; else c = 'g'; end;
scatter(ret(is), is, [], c, 'MarkerFaceColor', c);
h = text(ret(is) + offset_x, is + offset_y, {labels{is}});
set(h, 'Color', c);
grid on;
xlabel('return / util'); ylabel('seq'); title('Blue: partially greater // Green: dominated // Red: undominated (expanded)');
end
