% function [theta, delta] = linpend_fuzzyqiter
%  Fuzzy Q-iteration for linearized pendulum stabilization in 0 state.
%  Learns Q(alpha_1 + alpha_2
%   [THETA, DELTA] = FUZZYQITER
%
% Parameters:
%
% Returns:
%   THETA   - computed parameter vector 
%   DELTA   - the evolution of the infinity-norm difference over subsequent parameter vectors

model = linpend_problem('model');
cfg = linpend_problem('fuzzy');

% pick up fuzzy centers and compute flat action quant
c = cfg.xgrids; 
cfg.U = flat(cfg.ugrids);
% shorthand variables
p = model.p;
n = model.n;
ni = model.ni;
% compute dimension variables
Nn = zeros(n, 1);
for i = 1:n, Nn(i) = length(c{i}); end;
N = prod(Nn);
M = size(cfg.U, 2);
roll = 0 * Nn;      % no rollover
tab = dec2base(0:2^n-1, 2) - 47; % helper for mdegs_p
if n > 1,
    mdegsfun = @mdegs_p;
else
    mdegsfun = @mdegs_1;
end;
    

disp('Computing MDP and activation data...');

% for each x and u, activation vector of f(x, u)
F = sparse(N * M, N);       % F = zeros(N * M, N), but sparse -- most of the mdegs will be 0
% reward rho(x, u)
R = zeros(N, M);

% iterate over (x, u) in (centers, discrete actions) and populate F, R matrices
cpoints = flat(c);  % matrix of center points, p on rows and point index on columns
x = zeros(p, 1);
for i = 1:N,
    for j = 1:M,
        % NOTE: BEGIN 
        % the code is generic for n interesting state variables except for this section
        x([1 2]) = cpoints(:, i) / 2;       % assume each angle half of the current "state"
        % some small random values on the speeds - control randomness by the scaling factor /20
        x([3 4]) = (2*rand(2, 1) - 1) .* model.maxx([3 4]) / 20;
        % apply f, rho, and store results
        [xplus R(i, j)] = feval(model.fun, model, x, cfg.U(:, j));
        [ind, mu] = feval(mdegsfun, xplus(1)+xplus(2), c, roll, Nn, n, tab);
        % END non-generic code
        F(i+(j-1)*N, ind) = mu;
    end;            % FOR over action discretization
end;                % FOR over state samples

% dispx('Performing linear Q-value iteration...', cfg.verb, BASEVERB);

% initialize linear parameter matrix
theta = zeros(N, M);

k = 1; conv = 0;
delta = nan(1, cfg.maxiter);
while ~conv && k <= cfg.maxiter,       % main loop

    thetaold = theta;
    theta = R + cfg.gamma * reshape(max(F * theta, [], 2), N, M);
    
    % infty-norm of vector representation of theta difference
    delta(k) = max(max(abs(theta - thetaold)));
    conv = delta(k) <= cfg.eps;
    k = k + 1;
end;        % while not converged and allowed more iterations

if conv,	flag = 0; disp(['Convergence detected after ' num2str(k-1) ' iterations. Algorithm stopped.']);
else        flag = 1; disp(['maxiter=' num2str(cfg.maxiter) ' exhausted. Algorithm stopped']);
end;

figure(10); clf;
surf(cfg.xgrids{1}, cfg.ugrids{1}, theta'); xlabel('x_1 + x_2'); ylabel('u'); title('Qhat^*(x_1+x_2, u)');

